#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::ContentBase;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Base;

use vars qw(@ISA @class_structure %class_index);
@ISA = qw(ExSite::Base);
@class_structure = ("section", "page", "content", "content_data");
%class_index = ("ExSite::Section"=>0, 
		"ExSite::Page"=>1, 
		"ExSite::Content"=>2, 
		"ExSite::ContentData"=>3, 
		"ExSite::ContentData2"=>3);

#------------------------------------------------------------------------

=pod

=head1 ContentBase.pm : generic content-management methods

These are tools that can be used/inherited by all CMS data types (ie.
sites, pages, content definitions, and content data).

=cut

#------------------------------------------------------------------------


# options
# name => content object name
# page => page this content is assigned to
# id => direct lookup of content version
# version => the version to select
# DB => database object (optional)

sub new ($%) {
    my($this, %option) = @_;

    # objectify object
    my $cobj = {};
    my $class = ref($this) || $this;
    bless $cobj, $class;

    # define database

    if (defined $option{DB}) {
	$cobj->{DB} = $option{DB};
    }
    else {
	$cobj->{DB} = $share{DB};
    }

    # setup content cache - in share, so that multiple objects can share it

    if (! exists $share{ccache}) {
	$share{ccache} = new ExSite::Cache;
    }

    # setup Base class features

    $cobj->initialize_object;

    # register CMS-specific handlers

#    $cobj->{DB}->handler("select_foreign_key",&CMS_select_foreign_key);

    # setup context hash

    $cobj->{context} = {};

    # end generic constructor; now run local class setup

    $cobj->setup(%option);
    return $cobj;
}

# set the context : needed to find() related content.
#
# Every data object in the CMS is part of a content stack:
# site -> page -> content-def -> content-data
#
# Setting the context consists of finding and loading the other objects 
# in this stack.  It is assumed that the current data object is already 
# loaded.
#
# mode is none|all|parent|child

sub set_context {
    my ($this,$mode,$version) = @_;
    return if $this->{context_set};
    $mode = $mode || "all";
    if ($mode eq "all") { $this->{context_set} = 1; }
    # start with current object
    my $class = ref $this;
    my $index = $class_index{$class};
    $this->{context}{$class_structure[$index]} = $this->{data};
    return if ($mode eq "none");
    if ($mode =~ /all|parent/) {
	# move up thru the content stack by following foreign key refs
	for (my $i = $index; $i > 0; $i--) {
	    my $table = $class_structure[$i];
	    my $ptable = $class_structure[$i-1];
	    my $pkey = $ptable."_id";
	    my $id = $this->{context}{$table}{$pkey};
	    if ($id) {
		$this->{context}{$ptable} = $this->fetch($ptable,$id);
	    }
	    else {
		$this->warn("no $table found");
		last;
	    }
	}
    }
    if ($mode =~ /all|child/) {
	# move down thru the content stack
	my @cstack = @class_structure;
	while (scalar @cstack && $cstack[0] ne 
	       $class_structure[$class_index{$class}]) { 
	    shift @cstack; 
	};
	foreach my $ctype (@cstack) {
	    next if (ref $this->{context}{$ctype} eq "HASH"); # already loaded
	    if ($ctype eq "page") {
		$this->{context}{$ctype} = $this->find_page("index.%");
	    }
	    elsif ($ctype eq "content") {
		$this->{context}{$ctype} = $this->find("page","data");
	    }
	    elsif ($ctype eq "content_data") {
		$this->{context}{$ctype} = $this->find_content_data($version);
	    }
	}
    }
}

#------------------------------------------------------------------------

=pod

context testing routines

=cut

#------------------------------------------------------------------------

sub is {
    my ($this,$type) = @_;
    # return true if this object is a $type
    return (ref $this =~ /^$type$/i);
}

sub has {
    my ($this,$type) = @_;
    # return true if this object has a $type loaded in its context
    return defined $this->{context}{$type};
}

sub has_content {
    my ($this) = shift;
    # return true if this has content data defined 
    return $this->has("content_data");
}

sub get { 
    my ($this,$type,$attribute) = @_;
    if ($type) {
	# default to ::Base behaviour if $type is not a context item
	return $this->{$type} if (ref $this->{context}{$type} ne "HASH");
	# if $type is a context item, return data for it
	return $attribute ? 
	    $this->{context}{$type}{$attribute} : 
	    $this->{context}{$type};
    }
    else {
	# no args - return this object's data
	return $this->{context}{$class_structure[$class_index{ref $this}]};
    }
}

sub id {
    my ($this,$type) = @_;
    my $type = $type ? $type : $class_structure[$class_index{ref $this}];
    return $this->{context}{$type}{$type."_id"};
}

#------------------------------------------------------------------------

=pod

=head2 find() - locate related or referenced content

find logic:
- look in current page's content objects
- look in templates' content objects, from nearest relation to furthest
- look in content libraries for this section
- look in content libraries for ancestral sections
- take first eligible match found

mode is "data" or "object"

=cut

#------------------------------------------------------------------------

sub find {
    my ($this,$cname,$mode) = @_;

    if (! $this->{context_set}) {
    #if (! $this->{context}{page}) {
	# missing our context - can't just check context_set flag, because
	# set_context calls us before the context is fully set.
	$this->set_context;
    }

    my $page = $this->{context}{page};
    return undef if (! $page);

    # search the current page
    my $c = $this->find_content($cname);

    # search the template stack
    while (! $c && $page->{template_id}) {
	$page = $this->fetch("page",$page->{template_id});
	$c = $this->find_content($cname,$page->{page_id});
    }

    # search the libraries
    if (! $c) { 
	# get current section
	my $section = $this->{context}{section};
	######## set the parent context first, to be sure it is defined?
	return undef if (! $section);
	my $sid = $section->{section_id};
	while ($sid) {
	    # get all libraries in this section
	    my @libs = $this->fetch_match("page",{type=>"library",
						  section_id=>$sid});
	    if (@libs > 0) {
		# search each library for a matching content object
		my $lib = shift @libs;
		while ($lib && ! $c) {
		    $c = $this->find_content($cname,$lib->{page_id});
		    $lib = shift @libs;
		}
		last if ($c);
	    }
	    $sid = $section->{parent_id};
	}
    }

    return $c if ($mode eq "data");
    require ExSite::Content;
    return new ExSite::Content(content=>$c);
}

# more find tools

sub find_page {
    my ($this,$name,$sid) = @_;
    return undef if (! $this->{context}{section});
    if (! $sid) { $sid = $this->{context}{section}{section_id}; }
    return undef if (! $sid);
    my @page = $this->fetch_match("page",{filename=>$name,section_id=>$sid},
				  "page_id");
    return $page[-1];
}

sub find_content {
    my ($this,$name,$pid) = @_;
    return undef if (! $this->{context}{page});
    if (! $pid) { $pid = $this->{context}{page}{page_id}; }
    return undef if (! $pid);
    my @content = $this->fetch_match("content",{name=>$name,page_id=>$pid},
				     "content_id");
    return $content[-1];
}

sub find_content_data {
    my ($this,$type,$cid) = @_;
    return undef if (! $this->{context}{content});
    if (! $cid) { $cid = $this->{context}{content}{content_id}; }
    return undef if (! $cid);
    my @content_data = $this->fetch_child("content_data",
					  "content",
					  $cid,
					  "content_data_id");
    if ($type =~ /current|active/) {
	# return most recent approved version
	for (my $ic = $#content_data; $ic >= 0; $ic--) {
	    # if status flag is set, this content is waiting for approval
	    if (! $content_data[$ic]->{status}) {
		return $content_data[$ic];
	    }
	}
	return undef;
    }
    else {
	# return most recent version
	return $content_data[-1];
    }
}


#------------------------------------------------------------------------

=pod

=head2 expand(): Content Expansion/Substitution

Expands the content in the current context, by replacing CMS tags with
the URLs or the raw content that they represent.

Options:

content => a hash of content names and values to use preferentially over
           those in the database
method => a combination of data,url,content,page,module, depending on 
          which expansions you want to apply to this content.  All are 
          selected, by default.
cms => if true, include CMS tools in the output

=cut

#------------------------------------------------------------------------

sub substitute {
    return &expand(@_);
}

sub expand { 
    my ($this,%option) = @_;
    return $this->{html} if ($this->{expanded} || $this->{noexpand});
    $this->info("expanding html");
    $this->{html} = $this->get_html($option{mode});
    if (! exists $option{content}) { $option{content} = {}; }
    my $method = $option{method} || "data,url,content,page,module"; 
    my $cmsmenu = 0;

    # Keep track of number of HTML substutitions, since they require 
    # repeating the substitution loop.
    my $htmlsub;

    do {

	# clear substitution counter
	$htmlsub = 0;

	# 1) page data substitutions, eg. <!--$name-->

	if ($method =~ /data/) { 
	    $this->{html} =~ s/<!--\$(\w+)-->/$this->{context}{page}{$1}/gs;
	}

	# 2) content-URL substitutions, eg. SRC=[[content-object-name]]

	if ($method =~ /url/) {
	    $this->url_substitute("[[","]]",%option);
	}

	# 3) page-URL substitutions, eg. HREF={{page-name/id}}

	if ($method =~ /page/) { 
	    $this->url_substitute("{{","}}");
	}

	# 4) HTML content substitutions, eg. <!--&content(content-obj-name)-->

	my ($start,$end,$cobj,$cobjname,$csub,$newtext,@cobjname);
	if ($method =~ /content/) { 
	    $start = index($this->{html},"<!--&content(");
	    while ($start != -1) {
		$htmlsub++;
		$end = index($this->{html},")-->",$start);
		@cobjname = split(/,/,substr($this->{html},$start+13,$end-$start-13));
		# first content object in list is substituted directly
		$cobjname = shift @cobjname;
		#cobjname = substr($this->{html},$start+13,$end-$start-13);

		if ($method =~ /dummy-content/) {
		    my $cont = join(",",$cobjname,@cobjname);
		    $cont =~ s/\"/%22/g;
		    $cont =~ s/</%3C/g;
		    $cont =~ s/>/%3E/g;
		    $newtext = "<img alt=\"content($cont)\" src=$config{server}{HTMLpath}/_ExSite/images/checker.png title=\"$cont\" name=\"$cont\" HEIGHT=50 WIDTH=100>";
		}
		elsif ($option{content}->{$cobjname}) {
		    # predefined content object
		    $newtext = $option{content}->{$cobjname};
		}
		else {
		    # look for matching content object in DB
		    $cobj = $this->find($cobjname);
		    $newtext = $cobj->get_html;
		    if ($option{cms} && 
			$cobj->{context}{content}{page_id} == 
			$this->{context}{page}{page_id}) {
			###### should allow hotlinks to undefined content
			###### objects that are being inherited from template
			# can only manage content tied directly to this page
			# to manage templates, view the template in cms mode
			$newtext .= $this->edit_content_link($cobj);
			# highlight editable content sections
			$newtext = "<div style=\"border-width:1px; border-color:red; border-style:dashed; padding:0px;\">$newtext</div>\n";
		    }
		}

		# remaining content objects in list are content objects to 
		# insert into the first content object
		
		my $icobj = 1;
		foreach $cobjname (@cobjname) {
		    if ($cobjname =~ /^\".+\"$/) {
			# quoted string - plaintext substitution
			# simple strings only - (no commas!)
			$cobjname =~ s/^\"//;
			$cobjname =~ s/\"$//;
			$newtext =~ s/<!--\&content\($icobj\)-->/$cobjname/g;
		    }
		    else {
			# another content object
			$newtext =~ s/<!--\&content\($icobj\)-->/<!--&content($cobjname)-->/g;
		    }
		    $icobj++
		}

		$this->{html} = substr($this->{html},0,$start).$newtext.substr($this->{html},$end+4);
		# check if there are any more HTML content substitutions to do
		$start = index($this->{html},"<!--&content(");
	    }
	}
	    
	# 5) Dynamic Content Driver substitutions, eg. <!--&module.method(args)-->
	if ($method =~ /module/) {
	    my (%dcd,$endname,$module,$modname,$modmethod,$modargs);

	    # add a CMS menu
	    if ($option{cms} && ! $cmsmenu) {
		$this->{html} =~ s/(<body[^>]*>)/$1\n<!--\&CMS\(menu\)-->/i;
		# ensure that cmsmenu is only inserted once
		$cmsmenu = 1;
	    }

	    $start = index($this->{html},"<!--&");
	    while ($start != -1) {
		# first, check to make sure this is really a DCD call
		if (substr($this->{html},$start,40) !~ /<\!--\&\w+(\.\w+)?\(/) {
		    # hm, we have some strange tags that are spoofing a 
		    # DCD call - perhaps a commented HTML escape character?
		    # We have to change the encoding slightly so it escapes
		    # the attention of ExSite...
		    $this->warn("confusing tags resembling DCD call are being modified");
		    substr $this->{html}, $start, 5, "<!-- &"; 
		    $start = index($this->{html},"<!--&");
		    next;
		}
		$htmlsub++;
		$endname = index($this->{html},"(",$start);
		$module = substr($this->{html},$start+5,$endname-$start-5);
		($modname,$modmethod) = split(/\./,$module,2);
		if (!$modmethod) { $modmethod = "write"; }
		$end = index($this->{html},")-->",$endname);
		$modargs = substr($this->{html},$endname+1,$end-$endname-1);
		
		# run the DCD
		
		if ($method =~ /dummy-module/) {
		    $newtext = "<img alt=\"$module($modargs)\" src=$config{server}{HTMLpath}/_ExSite/images/checker.png title=\"$module($modargs)\" name=\"$module($modargs)\" HEIGHT=50 WIDTH=100>";
		}
		elsif ($modmethod =~ /^(read|write|ioctl)$/) {
		    my $stat = eval "require Modules::$modname";
		    if ($stat) {
			if (! $dcd{$modname}) {
			    # create the DC object
			    $dcd{$modname} = eval "new Modules::$modname";
			}
			# execute the DC method
			$newtext = $dcd{$modname}->$modmethod($modargs) || $@;
			# substitute the results back into the page
			if ($modname eq "ioctl") {
			    # ioctl results reported as diagnostic comments
			    # (can also return other DCD methods)
			    $newtext = "<!--$newtext-->";
			}
			elsif ($modname eq "read") {
			    # (no substitutions occur on reads)
			    $newtext = "";
			}
		    }
		    else {
			# $newtext = "<!-- Error in $modname dynamic content driver -->";
			# $@ is $EVAL_ERROR
			$newtext = "<H4>Error in $modname dynamic content driver:</H4>\n<P>\n$@\n</P>";
			$this->error("$modname: $@");
		    }
		}
		else {
		    $newtext = "<!-- Error: invalid dynamic content driver method: $modmethod -->";
		    $this->error("invalid DCD method: $modmethod");
		}
		$this->{html} = substr($this->{html},0,$start).
		    $newtext.
		    substr($this->{html},$end+4);
		
		$start = index($this->{html},"<!--&");
	    }
	}
    } until ($htmlsub == 0);

    $this->{expanded} = 1;  # flag expansion as done
    return $this->{html};
}

# url_substitute : replace content references in [[...]] or {{...}} with a URL
# $starttext = text denoting start of content reference, eg "[["
# $endtext   = text denoting end   of content reference, eg "]]"

sub url_substitute {
    my ($this,$starttext,$endtext,%option) = @_;
    my $len = length $starttext;
    my $start = index($this->{html},$starttext);
    my ($vfile,$subobj);
    # loop over every content reference with these delimiters in the page data
    while ($start != -1) {
	$vfile = "";
	my $end = index($this->{html},$endtext,$start);
	my $subname = substr($this->{html},$start+$len,$end-$start-$len);
	if ($endtext eq "}}") {
	    my $p;
	    if ($subname =~ /^\d+$/) {
		# page id
		$p = new ExSite::Page(id=>$subname,noload=>1);
	    }
	    else {
		# page name
		$p = new ExSite::Page(filename=>$subname,
				      section_id=>$this->id("section"));
	    }
	    $vfile = $p->get_url;
	}
	else {
	    # content reference
	    $subobj = $this->find($subname);
	    if (! $subobj || ! $subobj->defined) {
		# no matching content - escape the tag chars in case this
		# is not a CMS tag
		$vfile = &html_escape($starttext.$subname.$endtext,"[]");
	    }
	    else {
		$vfile = $subobj->get_url($option{mode});
	    }
	}
	substr($this->{html},$start,(length($starttext) + 
				     length($subname) + 
				     length($endtext))) = $vfile;
	# add CMS controls
	if ($endtext ne "}}" && 
	    $option{cms} && 
	    $subobj->{context}{content}{page_id} == 
	    $this->{context}{page}{page_id}) {
	    # can only manage content tied directly to this page
	    # to manage template content, view the template in cms mode
	    my $edittagpos = index($this->{html},">",$start);
	    my $edittags = $this->edit_content_link($subobj);
	    substr($this->{html},$edittagpos+1,0) = $edittags;
	}
	# on to next content object...
	$start = index($this->{html},$starttext);
    }
}


# wrappers for the DB data fetching methods, so that we can cache and reuse
# content fetched already

### TODO: additonal fetching/caching performance improvements can be
### realized by prefetching all content in advance, rather than doing
### individual searches for each content object.

sub fetch {
    my $this = shift;
    my $dat = $share{ccache}->get(@_);
    if (! $dat) {
	$dat = $this->{DB}->fetch(@_);
	$share{ccache}->save($dat,@_);
    }
    if (wantarray) { return %$dat; }
    return $dat;
}

sub fetch_child {
    my $this = shift;
    my $dat = $share{ccache}->get(@_);
    if (! $dat) {
	$dat = $this->{DB}->fetch_child(@_);
	# cache the whole query
	$share{ccache}->save($dat,@_);
	# now cache the individual records
	#my $pkey = $this->{DB}->get_key($_[0]);  # calls DB server
	my $pkey = $this->{DB}{map}->get_table_attr($_[0],"key"); # no server call
	foreach my $row (@$dat) {
	    # cache under same index as a simple fetch
	    $share{ccache}->save($row,$_[0],$row->{$pkey});
	}
    }
    if (wantarray) { return @$dat; }
    return $dat;
}

sub fetch_match {
    my ($this,$table,$match,$sort) = @_;
    # expand the match hash for best indexing
    # match can be a scalar or hash ref
    my @match = ref $match ? %$match : ($match);
    my $dat = $share{ccache}->get($table,@match,$sort);
    if (! $dat) {
	$dat = $this->{DB}->fetch_match($table,$match,$sort);
	$share{ccache}->save($dat,$table,@match,$sort);
	# now cache the individual records
	#my $pkey = $this->{DB}->get_key($table); # calls DB server
	my $pkey = $this->{DB}{map}->get_table_attr($table,"key"); # no server call
	foreach my $row (@$dat) {
	    # cache under same index as a simple fetch
	    $share{ccache}->save($row,$table,$row->{$pkey});
	}
    }
    if (wantarray) { return @$dat; }
    return $dat;
}

# get_dcd : return an array of DCD names known to this implementation

sub get_dcd {
    my $this = shift;
    my @dcd;
    ### $config{server}{CGIroot} is incorrectly defined
    my $CGIroot = $ENV{SCRIPT_FILENAME};
    $CGIroot =~ s/\/\w*\.cgi.*//;
    opendir MODDIR, "$CGIroot/Modules";
    my ($mod,$napp);
    @dcd = readdir MODDIR;
    closedir MODDIR;
    @dcd = grep /\.pm$/, @dcd; # only accept Perl modules
    @dcd = grep !/^BaseDCD\.pm$/, @dcd; # discard BaseDCD
    return sort @dcd;
}

#------------------------------------------------------------------------
#
# placeholder methods : these are dummies that will be overloaded in
# higher-level CMS classes.  They are provided here to allow ContentBase
# to run stand-alone for testing.

sub get_raw {
    return;
}

sub get_html {
    return;
}

sub get_url {
    return;
}

sub setup {
    return;
}

sub edit_content_link {
    return;
}
    

1;
