#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::ContentData;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ContentBase;
use ExSite::Cache;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::ContentBase);

#------------------------------------------------------------------------

=pod

=head1 ContentData.pm : tools for manipulating content data 

A ContentData object is a particular piece of raw web content that is
attached to a content object.  This modules provides methods to act on
ContentData objects.

new: define a new ContentData object

content_type: return the mime type of the content data

get: return the object in the specified mode

get_raw: return the object's raw data

get_html: return the HTML tags needed to inline the object into a page

get_url: return the URL needed to fetch the object in a separate request

=cut

#------------------------------------------------------------------------


# setup info for ContentData class

sub setup {
    &setupContentData(@_);
}

# fetch the content_data record

sub setupContentData {
    my ($this,%option) = @_;
    my $cdata;
    if ($option{id}) {
	$cdata = $this->fetch("content_data",$option{id});
    }
    elsif ($option{content_data}) {
	$cdata = $option{content_data};
    }
    if (ref $cdata eq "HASH") {
	$this->info("setting up ContentData $cdata->{content_data_id}");
	$this->{data} = $cdata;
	$this->set_context("none");
    }
    else {
	$this->warn("setting up null ContentData");
    }
}

# is this content defined?

sub defined {
    my $this = shift;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	return ($data->{data} || $data->{fdata} || 
		$data->{file} || $data->{url});
    }
    return 0;
}
	
# content queries

# return the content-type

sub content_type {
    my $this = shift;
    return $this->get("content_data","mime_type");
}

# determine whether content is in a virtual file

sub is_file {
    my $this = shift;
    my $dat = $this->get("content_data");
    return ( ! $dat->{data} && $dat->{fdata} );
}

# determine whether content is an image

sub is_image {
    my $this = shift;
    return $this->get("content_data","mime_type") =~ /^image\//;
}

# determine whether content is text

sub is_text {
    my $this = shift;
    return $this->get("content_data","mime_type") =~ /^text\//;
}

# determine whether content is publishable to a file
# -> essentially true for all non-HTML mime types

sub publishable {
    my $this = shift;
    my $mimetype = $this->get("content_data","mime_type");
    return ($mimetype && $mimetype !~ /html/);
}

# return the raw content

##### check for published file, file pointer, DB file, and raw content

sub get_raw {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{data}) {
	    return $data->{data};
	}
	elsif ($data->{fdata}) {
	    my ($file,$fdata) = split /$config{form}{sepchar}/,$data->{fdata};
	    return decode_base64($fdata);
	}
    }
    $this->warn("ContentData: no raw content defined");
    return undef;
}

# return the HTML tags needed to inline the content into a page

sub get_html {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	my $type = $this->content_type;
	if ($type =~ /^text\/html$/) {
	    # html can be returned straight-up
	    return $this->get_raw;
	    #return $this->substitute;  ##### weird error on page views here
	}
	elsif ($type =~ /^text\/css$/) {
	    # stylesheets are referenced using a link tag
	    my $url = $this->get_url;
	    return "<link rel=\"stylesheet\" type=\"text/css\" href=\"$url\">";
	}
	elsif ($type =~ /^text/) {
	    # plain text is rendered preformatted
	    return "<pre>\n".$this->get_raw."</pre>\n";
	    # alternately, return it as-is
	    #return $this->get_raw;
	}
	elsif ($type =~ /^image/) {
	    # images are rendered using IMG tags
	    my $url = $this->get_url(@_);
	    return "<img src=\"$url\" $data->{attributes}>";
	}
	else {
	    # miscellaneous types are shown as hotlinks
	    my $url = $this->get_url(@_);
	    my $filename = $this->filename || "download $type";
	    return "<a href=\"$url\" $data->{attributes}>$filename</a>";
	}
    }
    $this->warn("ContentData: no html content defined");
    return undef;
}

# return a URL that fetches the content as a separate HTTP request
# mode = static|dynamic (defaults to static, if found, dynamic otherwise)

sub get_url {
    my $this = shift;
    my $mode = shift;
    if ($this->{publish}) { $mode = undef; }
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	# if there is a published version, use it
	if ($mode ne "dynamic") {
	    my $file = $this->filename;
	    my $diskpath = $this->diskpath;
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		my $url;
		if ($data->{type} eq "page") {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    require ExSite::Page;
		    my $p = new ExSite::Page(page=>$data);
		    $url = $p->get_url;
		}

		## make sure we know our section
		#$this->set_context; 
		#require ExSite::Section;
		#my $sec = new ExSite::Section(section=>$this->get("section"));
		## content is saved in the section, presumably
		#if (-e "$config{server}{HTMLroot}$config{server}{HTMLpath}$url/$file") {
		### fails if $url contains server info
		    return $url."/".$file;
		#}
	    }
	}
	if ($mode ne "static") {
	    # use dynamic version
	    my $url = "$config{server}{server}$config{server}{CGIpath}/content.cgi?id=".$this->id("content_data");
	    my $cobj = $this->get("content");
	    $url .= "&name=$cobj->{name}"; 
	    return $url;
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}->{HTMLpath}/_ExSite/images/checker.png";
}

sub get_static_url {
    my $this = shift;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{content_data_id}) {
	    my $file = $this->filename;
	    ################# testme: is diskpath stuff needed?
	    ##### don't understand the diskpath/get_url/file logic here
	    my $diskpath = $this->diskpath;
	    #if (-e "$diskpath/$file") {
	    #print "looking for content data $data->{content_data_id} in $diskpath<br>\n";
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		my $url;
		if ($data->{type} eq "page") {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    require ExSite::Page;
		    my $p = new ExSite::Page(page=>$data);
		    $url = $p->get_url;
		}
		return $url."/".$file;
	    }
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}->{HTMLpath}/_ExSite/images/checker.png";
}

sub get_dynamic_url {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{content_data_id}) {
	    my $url = "$config{server}{server}$config{server}{CGIpath}/content.cgi?id=".$this->id("content_data");
	    my $cobj = $this->get("content");
	    $url .= "&name=$cobj->{name}"; 
	    return $url;
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}{server}$config{server}{HTMLpath}/_ExSite/images/checker.png";
}

### FIXME: do we need a get method that returns an administrative link to 
### the content data (or content object)?

# determine the filename of a content datum "virtual file"

sub filename {
    my $this = shift;
    my $file;
    my $data = $this->get("content_data");;
    if (ref $data eq "HASH") {
	if ($data->{data}) {
	    $file = $data->{content_id};   # more constant than content_data_id
	    my $mtype = $data->{mime_type};
	    if ($mtype =~ /text\/html/) {
		$file .= ".html";
	    }
	    elsif ($mtype =~ /text\/css/) {
		$file .= ".css";
	    }
	    elsif ($mtype =~ /^text/) {
		$file .= ".txt";
	    }
	    ######## FIXME: need a mimetype to suffix converter
	}
	elsif ($data->{fdata}) {
	    ($file, undef) = split /$config{form}{sepchar}/,$data->{fdata};
	    $file =~ s/ /_/g;
	}
	return $file;
    }
    return undef;
}

# determine the path to the file, if published

sub httppath {
    my $this = shift;
    $this->set_context;
    require ExSite::Page;
    my $page = new ExSite::Page(page=>$this->get("page"));
    # the path to the content data is the same as the path to its page
    return $page->httppath;
}

# determine the path to the file, if published

sub diskpath {
    my $this = shift;
    $this->set_context;
    #return undef if (! $this->{cdata});
    #return $config{server}->{HTMLroot}.$this->httppath;
    my $data = $this->get("page");
    my $path;
    if ($data->{type} eq "page") {
	# page files are written directly to the section directory
	require ExSite::Section;
	my $sec = new ExSite::Section(section=>$this->get("section"));
	$path = $sec->diskpath;
    }
    else {
	# page files are written to the library/template subdirectory
	require ExSite::Page;
	my $p = new ExSite::Page(page=>$data);
	$path = $p->diskpath;
    }
    $path .= "/".$this->filename;
    return $path;
}

# preview: display the content for an administrator

sub preview {
    my $this = shift;
    my $type = $this->content_type;
    if ($type eq "text/html") {
	# expand the content
	return $this->expand(mode=>"dynamic",method=>"data,url,content,page");
    }
    elsif ($type =~ /^text/) {
	# display raw content for inspection, rather than convert it to HTML
	return "<pre>\n".$this->get_raw."\n</pre?\n";
    }
    else {
	# convert to HTML - force dynamic preview mode for images
	return $this->get_html("dynamic");
    }
}

# write the content data out to a file

sub publish {
    my ($this,$path,$filename) = @_;
    if ($path) {
	my $data = $this->get("content_data");
	if (ref $data eq "HASH") {
	    my ($file,$fdata);
	    $file = $filename || $this->filename;
	    $fdata = $this->get_raw;
	    my $outputpath = $path ? "$path/$file" : $file;
	    $outputpath =~ /^(((\/[\w-]+)+)\/[\w-]+\.\w+)$/;
	    my $secure_outputpath = $1;
	    $this->info("publish content data $data->{content_data_id} to $secure_outputpath");
	    if (open OUTFILE, ">$secure_outputpath") {
		print OUTFILE $fdata;
		close OUTFILE;
		return "Successfully published $file.<br>\n";
	    }
	    else {
		$this->error("Failed to publish $file: $!");
		my $x = $this->get("content_data");
		return "<p class=error>Failed to publish $file: $!</p>\n";
	    }
	}
    }
    $this->warn("ContentData::publish: No publication directory specified");
    return undef;
}


1;
