package ExSite::DB;
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

#======================================================================

=pod

=head1 ExSite::DB - generic database driver wrapper for ExSite

The DB class is a wrapper class for one of the db driver classes.
It inherits general query methods from one, but only one of these,
determined at runtime.

DB also provides an assortment of simplified and/or convenience
wrapper methods for accessing the low-level database driver routines.

In addition, the DB class provides an assortment of security,
access control, sorting, and row relationship methods.

The DB class is not normally used directly, unless the program is
doing pure data analysis.  Normally the Form:: and Report:: classes
are preferred for web application work, as these classes inherit all
of DB's methods.

=cut

#======================================================================

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::SQL;
#use ExSite::DBM;   not working, anyway
use ExSite::Text;
use ExSite::DBmap;
use MIME::Base64;

use vars qw(@ISA %cplog);

#======================================================================

=pod

=head2 new(%options) - DB constructor

new() creates a database handle.  The following arguments should be 
specified in the options:

=over 4

=item * type = the database type ("sql" and "text" are most common types)

=item * name = database name

=item * map = map name

=item * any other args required by driver (eg. "user" and "pass" may
be required to access SQL databases).

=back

Note that the map is a small database that describes the main database.
The map name is the name of the map database - it will be opened up
using the DBmap class.

=cut

#======================================================================

sub new {
  my ($this, %opt) = @_;
  my $obj = {};
  my $class = ref($this) || $this;
  bless $obj, $class;
  # if no options are provided, use default database hash
  my @opkeys = keys %opt;
  if (@opkeys == 0) { %opt = %{$config{server}->{db}}; }
  # setup database
  # inheritance is determined at runtime (!)
  # must copy relevant attributes from our temporary db object into
  # the wrapper object, so that inherited methods are duped into
  # thinking they are working on an object that they created.
  $_ = $opt{type};
  my $db;
  if (/sql/){
    $db = new ExSite::SQL(%opt);
    # inherit from sql
    push @ISA, "ExSite::SQL";
    $obj->{dbtype}= "ExSite::SQL";
  }
#  elsif (/dbm/) {
#    $db = new ExSite::DBM(%opt);
#    # inherit from dbm
#    push @ISA, "ExSite::DBM";
#  }
  elsif (/text/) {
    $db = new ExSite::Text(%opt);
    # inherit from text
    push @ISA, "ExSite::Text";
    $obj->{dbtype} = "ExSite::Text";
  }
  else {
    $obj->error("ExSite::DB ERROR: unknown database type ($_)");
    return undef;
  }

  # copy attributes of whatever db type we've created into our db object
  foreach my $key (keys %$db) {
    $obj->{$key} = $db->{$key};
  }
  #$obj->{db} = $db;  ###?

  # setup dbmap
  my $map;
  if ($opt{map}) {
      $map = new ExSite::DBmap( name=> $config{server}->{$opt{map}}->{name} );
      #$map = new ExSite::DBmap( %{$config{server}->{map}} );
  }
  else {
      #$map = new ExSite::DBmap( name=> undef );
      $map = new ExSite::DBmap;
      # no explicit list of tables, no real database to look up, so we must
      # manually set the list of mappable tables.
      $map->set_mapped_tables($db->get_tables);
  }

  if (! $map) {
      $obj->error("DB ERROR: no dbmap");
      return undef;
  }
  $obj->{map} = $map;

  # authorization type ("cookie", "fixed", or "none")
  # (or you can install your own handlers to do the authorization)
  $obj->{authentication} = $opt{authentication} || $config{auth}->{type};

  # error accumulators - used for user display errors;
  # For loggable system errors, use error() method instead.
  $obj->{errors} = [];      # fatal errors - changes will not be committed

  # store a pointer to the DB object in %share, unless save is disabled
  if (exists $opt{save} ? $opt{save} : 1) {
      $share{DB} = $obj;
  }

  # setup Base class features
  $obj->initialize_object;

  # handlers - done in base class now
  #&install_handlers($obj);

  # cache for single-record lookups
  $obj->{cache} = {};

  return $obj;
}

#======================================================================

=pod

=head1 High-level DB access methods

These methods invoke the low-level methods select, insert, update, and
delete, but otherwise have no knowledge of the underlying driver.
Unlike the low-level driver methods, they can also make use of the
dbmap, if needed.

All fetch methods return datahashes, that is one hash per record with
keys equal to the field names, and values equal to the data for those
fields.  All fields in each record are returned, by default, including
fields that the user is not permitted to access (access control is
performed at a higher level, in Form:: or Report::).

The fetch methods will take an optional SORT argument, which will
attempt to get the database driver to sort the resulting records
according to the value of SORT, which can either be a field name, or a
reference to a list of field names.  If no SORT argument is specified,
the data will be unsorted, but may still be sorted using the
DB::sort_rows() method.

=head2 fetch_key() - fetch a record by its primary key

Usage: C<$db-E<gt>fetch_key(TABLE, KEY, SORT)>

Returns the single record from C<TABLE> with primary key C<KEY>.  If
more than one record has that primary key, this routine returns the
first one found.  SORT is optional.

The C<fetch()> method is synonymous with C<fetch_key()>.

=head2 fetch_match() - fetch a record using a match hash.

Usage: C<$db-E<gt>fetch_match(TABLE, MATCH, SORT)>

MATCH is a reference to a hash containing field =E<gt> value pairs.
The values are assumed to be text segments that will be used to match
data in the actual database.  The SQL "%" wildcard may be used.
Records that match ALL field =E<gt> value pairs will be returned.
SORT is optional.

=head2 fetch_all() - fetch all records in a table

Usage: C<$db-E<gt>fetch_all(TABLE, SORT)>

Returns all records in the table, optionally sorted according to the
value of SORT.

=head2 fetch_child() - fetch related rows in a sub-table

Usage: C<$db-E<gt>fetch_child(TABLE, PARENT, PARENT_KEY, SORT)>

C<fetch_child()> is a trivial join, fetching all rows from table B
that relate to a single row in table A, where table B references A
using its primary key.  The rows in B are "children" of the row in A.
(There is no need to use a real JOIN mechanism, since we already know
which row in A we are matching to.)  SORT is optional.

=head2 fetch_join() - fetch related rows in a sub-table

Usage: C<$db-E<gt>fetch_join(TABLE_LIST, MATCH_LIST)>

In development.

=head2 fetch_m2m() - fetch rows that are related via a join table.

Usage: C<$db-E<gt>fetch_m2m(TABLE_LIST, MATCH_LIST)>

Say that table A relates to table B via a junction table J that contains
references to the primary keys of both A and B.  fetch_m2m will
perform a join on A,B,J, selecting all members of B that
relate to a specific member of A via the junction table J.

Example:  a singer can have songs on many albums, and an album can
contain many different singers.  If we have tables "singer", "album",
and "singer_album" (which contains the primary keys of particular
singer/album pairs), then we can select all the albums with songs by
singer "XYZ" using:

C<$db-E<gt>fetch_m2m("album","singer","XYZ","singer_album");>

(read this as "fetch albums for singer XYZ using singer_album")

=cut

#======================================================================

#----------------------------------------------------------------------
# fetch_key : fetch a row by its primary key

sub fetch_key {
  my ($this,$table,$key,$sort) = @_;
  # get the true name of the table, and the primary key name
  my $tab = $this->{map}->get_table_name($table);
  #my $pkey = $this->get_key($tab);  # calls DB server
  my $pkey = $this->{map}->get_table_attr($tab,"key");  # no server call
  my @dat;
  # check the cache first
  if (exists $this->{cache}{$tab}) {
      if ($this->{cache}{$tab}{$key}) {
	  # this record has already been fetched
	  @dat = ($this->{cache}{$tab}{$key});
      }
  }
  else {
      $this->{cache}{$tab} = {};
  }
  if (@dat == 0) {
      # look up the record
      @dat = $this->select($tab,undef,{$pkey=>$key},$sort);
      # cache the result
      $this->{cache}{$tab}{$key} = $dat[0];
  }
  # return only one record
  if (wantarray && @dat > 0) { return %{$dat[0]}; }
  return $dat[0];
}

# fetch is synonymous for fetch_key
sub fetch {
  return &fetch_key(@_);
}

#----------------------------------------------------------------------
# fetch_match : fetch rows using a match hash

sub fetch_match {
  my ($this,$table,$match,$sort) = @_;
  my $tab = $this->{map}->get_table_name($table);
  if ($tab ne $table) {
      # this is an aliased table - check for fixed fields
      my $ncol = $this->{map}->get_ncol($table);
      for (my $icol = 0; $icol <= $ncol; $icol++) {
	  if ($this->{map}->get($table,$icol,"datatype") eq "fixed") {
	      $match->{$this->{map}->get($table,$icol,"column")} = 
		  $this->{map}->get($table,$icol,"default");
	  }
      }
  }
  return $this->select($tab,undef,$match,$sort);
}

#----------------------------------------------------------------------
# fetch_all : fetch all available rows

sub fetch_all ($$$) {
    my ($this,$table,$sort) = @_;
    my $tab = $this->{map}->get_table_name($table);
    if ($tab ne $table) {
	# this is an aliased table - check for fixed fields
	my $match = {};
	my $ncol = $this->{map}->get_ncol($table);
	for (my $icol = 0; $icol <= $ncol; $icol++) {
	    if ($this->{map}->get($table,$icol,"datatype") eq "fixed") {
		$match->{$this->{map}->get($table,$icol,"column")} = 
		    $this->{map}->get($table,$icol,"default");
	    }
	}
	return $this->select($tab,undef,$match,$sort);
    }
    else {
	return $this->select($tab,undef,undef,$sort);
    }
}

#----------------------------------------------------------------------
# fetch_child : fetch descendant rows
#
# ...uses the DBmap to determine the reference keys.

sub fetch_child {
  my ($this,$table,$parent_table,$parent_id,$sort) = @_;
  # select rows referring to $parent_id
  my $key = $this->{map}->get_reference_column($table,$parent_table);
  my $tab = $this->{map}->get_table_name($table);
  return $this->select($tab,undef,{$key=>$parent_id},$sort);
}

=pod
# scratched this - unworkable
#----------------------------------------------------------------------
# fetch_join : fetch from a set of joined tables
#
# Parameters:
# - reference to a list of tables
# - reference to a list of match hashes
#

sub fetch_join {
  my ($this,$rtable,$rmatch) = @_;
  # combine all the match hashes
  my @matches = @$rmatch;
  my %match;
  foreach my $table (@$rtable) {
      my $tab = $this->{map}->get_table_name($table);
      my $newmatch = shift @matches;
      while (my ($key,$val) = each %$newmatch) {
	  $match{"$tab.$key"} = $val;
      }
  }
  # select all columns from all tables
  return $this->select($rtable,undef,\%match);
}
=cut

#----------------------------------------------------------------------
# fetch_m2m : fetch rows via a junction table
#
# ...uses the DBmap to determine the reference keys.

sub fetch_m2m {
  my ($this,$table1,$table2,$key2,$jointable,$sort) = @_;
  my ($ref1,$ref2,$pkey1,$pkey2,@columns);
  my $tab1 = $this->{map}->get_table_name($table1);
  my $tab2 = $this->{map}->get_table_name($table2);
  my $jtab = $this->{map}->get_table_name($jointable);

  # find the join columns that reference the parent tables
  $ref1 = "$jtab.".$this->{map}->get_reference_column($jtab,$tab1);
  $ref2 = "$jtab.".$this->{map}->get_reference_column($jtab,$tab2);

  # get primary keys, with explicit table references added
  $pkey1 = "$tab1.".$this->get_key($tab1);
  $pkey2 = "$tab2.".$this->get_key($tab2);

  # we only want to select the columns from table1
  @columns = $this->get_columns($tab1);
  # add explicit table reference to column names
  foreach my $col (@columns) {
    $col = "$tab1.$col";
  }

  # In SQLese:
  # SELECT FROM A,B,J WHERE B.PKEY="key2" AND J.REF2="key2" AND J.REF1=A.PKEY
  return $this->select([ $tab1, $tab2, $jtab ],
		       \@columns,
		       { $pkey2=>$key2, $ref2=>$key2, $ref1=>\$pkey1 },
		       $sort);
}

  # example query:
  # select title from publication,member,author_link where member.member_id="440" AND author_link.member_id="440" AND author_link.publication_id=publication.publication_id;
  #return $this->fetch_custom("SELECT $column_list FROM $table1,$table2,$join_table WHERE $table2.$pkey2=\"$key2\" AND $join_table.$ref2=\"$key2\" AND $join_table.$ref1=$table1.$pkey1");


#----------------------------------------------------------------------
# delete_key : delete a row by its primary key

sub delete_key {
  my ($this,$table,$key) = @_;
  my $tab = $this->{map}->get_table_name($table);
  my $pkey = $this->get_key($tab);
  return $this->delete($tab,{$pkey=>$key});
}

#----------------------------------------------------------------------
# copy : copy a record
# if the recursive flag is true, will recusively copy all descendant records

#### DANGER: copying records that have multiple references to a parent table
#### will not be able to accurately copy the foreign key references.

# sub copy {
#     my ($this,$table,$rec,$recursive) = @_;
#     %cplog = ();            # initialize the infinite loop trap
#     my $id = &copy_1(@_);   # copy the head record
#     if ($recursive) {
# 	&copy_r(@_);   # copy descendant records
#     }
# }
# 
# # copy single record
# 
# sub copy_1 {
#     my ($this,$table,$rec) = @_;
#     my $data;
#     if (ref $rec) {
# 	$data = $rec; 
#     }
#     else {
# 	$data = $this->fetch($table,$rec);
#     }
#     my $key = $this->get_key($table);
#     my $keytype = $this->{map}->get($table,$key,"datatype");
#     if ($keytype = "key#") {
# 	delete $data->{$key};
#     }
#     my $id = $db->insert($table,$data);
#     $cplog{"$table:$rec"} = 1;
#     return $id;
# }
# 
# # recursive copy routine
# 
# sub copy_r {
#     my ($this,$table,$rec,$pcol,$newpid) = @_;
#     return if ($cplog("$table:$rec"));  # infinite loop trap
#     &copy_1(@_);
#     if ($recursive) {
# 	my @childtables = $this->{map}->get_children($table);
# 	foreach my $ctab (@childtable) {
# 	    my @child = $this->fetch_child($ctab,$table,$rec);
# 	    my $ckey = $this->get_key($ctab);
# 	    foreach my $crec (@child) {
# 		$this->copy_r($ctab,$crec->{$ckey},1);
# 	    }
# 	}
#     }
# }

#======================================================================
# wrappers for low-level database driver methods insert, update,
# delete, select.  Allows us to apply DBmap logic to our database
# requests - in particular, table aliases.

sub get_columns {
    my ($this,$table) = @_;
    my $dbtable = $this->{map}->get_table_name($table);
    return $this->dbget_columns($dbtable);
}

##### could fetch key from dbmap instead (faster!)
sub get_key {
    my ($this,$table) = @_;
    my $dbtable = $this->{map}->get_table_name($table);
    return $this->dbget_key($dbtable);
}

#### include security checks for access to tables, columns, rows

sub select {
    my $this = shift;
    my $table = shift;
    my $dbtable;
    # map DBmap table names to low-level table names
    if (ref $table eq "ARRAY") {
	# array of tables was passed
	my @tables;
	foreach my $t (@$table) {
	    push @tables, $this->{map}->get_table_name($t);
	}
	# replace $table with unaliased version
	$dbtable = \@tables;
    }
    else {
	# single table was passed as a scalar - unalias it
	$dbtable = $this->{map}->get_table_name($table);
    }
    return $this->dbselect($dbtable,@_);
}

sub insert {
    my $this = shift;
    my $table = shift;
    my $dbtable = $this->{map}->get_table_name($table);
    return $this->dbinsert($dbtable,@_);
}

sub update {
    my $this = shift;
    my $table = shift;
    my $dbtable = $this->{map}->get_table_name($table);
    return $this->dbupdate($dbtable,@_);
}

sub delete {
    my $this = shift;
    my $table = shift;
    my $dbtable = $this->{map}->get_table_name($table);
    return $this->dbdelete($dbtable,@_);
}

#======================================================================

=pod

=head1 Authorization

Authorization is performed before any standard operation on
the database, when the operation is performed by one of ExSite::DB's
descendant classes.  The authorization process has the following
steps:

=over 4

=item 1. identify - determine who the user claims to be

=item 2. lookup - determine whom we understand that username to represent

=item 3. authenticate - determine if the user's claim matches our knowledge

=item 4. authorize - grant a certain level of access to the database

=item 5. approve - approve a specific operation

=back

Each of the steps in the authorization process may be replaced with
a custom handler if the procedure differs from the default ExSite logic.

=head2 identify() - determine who the user claims to be

Determines the user's claimed identity (such as by checking the
identity cookie), and stores the result so that authorization methods
can check it.  This routine does not authenticate the user's identity
claim, and so is usually not called directly.

=head2 lookup() - look up the user to see if we know who they are

Consult our own database to see if we have a record corresponding to
this user's identity (determined from C<identify()>).  Returns the
user record in a hash, if found.  The default logic for looking up a
user record is to look up the login name, which implies that unique
login names are required.  Once this lookup has occurred, the
following identity query functions can be used:

=head2 my_name() - return the login name of the current user

=head2 my_uid() - return the primary key of the current user's record

=head2 my_gid() - return the primary key of the current user's group record

=head2 my_user_record() - return the user's identity hash

=head2 authenticate() - verify the user's identity claim

Returns true (1) if the user's identity matches the information
returned by lookup(), false (0) otherwise.  This serves as a simple
check that the user is logged in, and that their identity is valid.

=head2 authorize() - grant the user a specific level of access

Returns the user's access level.  Access level is a value of 0 or
higher.  0 corresponds to "nobody", or the general public.  Values of
1 or more correspond to authenticated users, with higher values giving
greater permissions to the database.  The maximum value is
C<$config{auth}-E<gt>{root_auth_level}>, which grants unrestricted
"super-user" access to the data.  (This level of access can be
dangerous, as it allows one to mess with sensitive data fields and
referential integrity.)

The default authorize routine assumes that the user's access level
is stored in the user table.  This value is simply returned without
any sanity checks.

=head2 approve() - approve a specific database operation

Usage: C<$db-E<gt>approve( OP, TABLE, RECORD );>

OP is a low-level database operation (one of "read", "update",
"insert", or "delete").  TABLE is the table that this operation is
being performed on. RECORD is the actual record datahash that this
operation is being performed on (only used for "update" and "delete").

The return value is true if the operation is permitted, or false if:

=over 4

=item * the user does not have permission to read or write to the table

=item * the user is referencing records owned by a different user or group.

=back

=cut 

#----------------------------------------------------------------------
# identify : determine whom the user claims to be

sub identify {
  my ($this) = shift;
  my ($name,$pwd,$uid,$gid);
  my $stat = $this->run_handler("identify");
  if (defined $stat) {
      # application has its own identify algorithm
      ($name,$pwd) = @$stat;
  }
  elsif ($this->{authentication} eq "cookie") {
      ($name,$pwd,undef) = &get_cookie($config{site}->{identity_cookie},":");
  }
  elsif ($this->{authentication} eq "fixed") {
      $name = $config{auth}->{default_identity}->{name};
  }
  elsif ($this->{authentication} eq "none") {
      $name = "nobody";
  }
  $this->{identity} = { password => $pwd };
  $this->{identity}->{name} = $name;     # name should be return value
}

#----------------------------------------------------------------------
# lookup : look up the user, to see if we know who they are

sub lookup {
  my ($this,$login) = @_;
  if ($this->{identity}->{lookup}) {
      return $this->{identity}->{lookup};
  }
  # check for custom lookup algorithm
  my $stat = $this->run_handler("lookup");
  return $stat if (defined $stat);
  #
  return undef if ($this->{authentication} eq "none");
  #my $udata = $this->fetch_match($config{auth}->{user_table},{$config{auth}->{user_login_column}=>$login});
  my $udata = $share{DB}->fetch_match($config{auth}->{user_table},{$config{auth}->{user_login_column}=>$login});
  if (@$udata > 0) {
      $this->{identity}->{lookup} = $udata->[0];
      $this->{identity}->{uid} = $udata->[0]->{$config{auth}->{user_uid_column}};
      $this->{identity}->{gid} = $udata->[0]->{$config{auth}->{user_gid_column}};
      $this->{identity}->{access} = $udata->[0]->{$config{auth}->{user_access_column}};
      return $udata->[0];
  }
  return undef;
}

# some identity queries, for convenience:

sub my_name {
  my ($this) = shift;
  if (exists $this->{identity}) {
      return $this->{identity}->{name};
  }
  return undef;
}

sub my_uid {
  my ($this) = shift;
  if (exists $this->{identity}) {
      return $this->{identity}->{uid};
  }
  return undef;
}

sub my_gid {
  my ($this) = shift;
  if (exists $this->{identity}) {
      return $this->{identity}->{gid};
  }
  return undef;
}

sub my_user_record {
  my ($this) = shift;
  if (exists $this->{identity}) {
      if (exists $this->{identity}->{lookup}) {
	  return %{$this->{identity}->{lookup}};
      }
      else {
	  return ();
      }
  }
  return undef;
}

#----------------------------------------------------------------------
# authenticate : verify the user's identity claim
#
# Returns true (1) if the user's identity matches the information
# returned by lookup(), false (0) otherwise.

sub authenticate {
  my ($this) = shift;
  if (exists $this->{authenticated}) {
    return $this->{authenticated};
  }
  # check for custom authentication algorithm
  my $stat = $this->run_handler("authenticate");
  return $stat if (defined $stat);
  if (exists $ENV{EXSITE_TEST} || $this->{authentication} eq "fixed") {
    $this->{authenticated} = 1;
    $this->{identity} = $config{auth}->{default_identity};
    return 1;
  }
  else {
    # implicit approval if authorization is turned off or $EXSITE_TEST is
    # set when testing from the shell
    return 1 if ($this->{authentication} eq "none");
    # check identity
    if (!$this->{identity}) { &identify($this); }
    if (!$this->{identity}->{name}) {
      # nobody logged in - not necessarily an error
      return 0;
    }
    # look up the user
    my $user = &lookup($this,$this->{identity}->{name});
    if (!$user) {
      # unknown username
      $this->warn("no such user \"$this->{identity}->{name}\"");
      return 0;
    }
    elsif (! $user->{$config{auth}->{user_pwd_column}}) {
      # password field is blank - deny access
      $this->warn("blank password for user \"$user->{$config{auth}->{user_login_column}}\"");
      return 0;
    }
    # compare entered and stored passwords
    my $pass;
    if ($config{auth}->{password_storage_method} eq "cleartext") {
      # encrypt the db password before comparing with identity
      my $salt = substr($this->{identity}->{password},0,2);
      $pass = ($this->{identity}->{password} eq
	       crypt($user->{$config{auth}->{user_pwd_column}},$salt));
    }
    else {
      # encrypted db passwd can be compared directly with identity
      $pass = ($this->{identity}->{password} eq $user->{$config{auth}->{user_pwd_column}});
    }
    if ($pass) {
      # passwords match
      $this->{authenticated} = 1;
      # store extra user info in identity
      $this->{identity}->{gid} = $user->{$config{auth}->{user_gid_column}};
      ####?
      $this->{identity}->{access} = $user->{$config{auth}->{user_access_column}};
      return 1;
    }
    else {
      # bad password
      $this->warn("incorrect password for user \"$user->{$config{auth}->{user_login_column}}\" - access denied");
      $this->{authenticated} = 0;
      return 0;
    }
  }
}

#----------------------------------------------------------------------
# authorize : grant the user a specific level of access
#
# Access level is a value of 0 or higher.  0 corresponds to "nobody",
# or the general public.  Values of 1 or more correspond to
# authenticated users, with higher values giving greater permissions
# to the database.  The maximum value is $config::root_auth_level,
# which grants unrestricted "super-user" access to the data.  (This
# level of access can be dangerous, as it allows one to mess with
# sensitive data fields and referential integrity.)
#
# The default authorize routine assumes that the user's access level
# is stored in the user table.  This value is simply returned without
# any sanity checks.

sub authorize {
  my ($this) = shift;
  my $stat = $this->run_handler("authorize");
  return $stat if (defined $stat);
  if (! exists $this->{authenticated}) {
      &authenticate($this);
  }
  if ($this->{authenticated}) {
      return $this->{identity}->{access};
  }
  # give public access by default
  return 0;
}

#----------------------------------------------------------------------
# approve : approve a specific database operation
#
# Arguments:
# op - "read", "update", "insert", "delete"
# table - the table being operated on
# [record] - reference to a datahash containing the record being
#            operated on (used for update, delete only)
#
# Returns true if the operation is permitted.
# Returns false if:
#   - the user does not have permission to read or write to the table
#   - the user is referencing records owned by a different user or group.
#   - the user does not have the required level of access for this CGI script

sub approve {
  my ($this,$op,$table,$record) = @_;
  my $stat = $this->run_handler("approve",$table,$record);
  return $stat if (defined $stat);
  my $level = $this->authorize();
  my $tablemap = $this->{map}->get_table($table);
#    # check script permissions first
#    split(/[\/\.]/,$ENV{SCRIPT_NAME});
#    my $cgiprog = $_[-2];
#    if ($level < $config{auth}->{cgi}->{$cgiprog}) {
#	return 0;
#    }
  # script okay, now check table permissions
  if (exists $tablemap->{$op}) {
      if ($op =~ /read|insert|search/) {
	  return ($level >= $tablemap->{$op});
      }
      else {
	  # for updates & deletions, must check row ownership
	  if (!$record) {
	      if ($level >= $config{auth}->{root_auth_level} - 1) { return 1; }
	      else { return 0; }
	  }
	  elsif ($this->user_owns($table,$record)) {
	      return ($level >= $tablemap->{$op});
	  }
	  else {
	      return 0;
	  }
      }
  }
  elsif ($this->{dbmap}) {
      # dbmap kludge - no tablemap or authentication when editing dbmaps
      return 1;
  }
  else {
      # unknown operation
      return 0;
  }
}

=pod

=head1 Ownership

The ExSite default user model has 5 levels of access: 0=public,
1=user, 2=group leader, 3=administrator, 4=root/super-user.

Rows can be owned by groups and by users.  The following methods are
used for determining row ownership for purposes of approving
operations (see above).

Ownership is determined by tracing data relationships until a
user or group record is arrived at.  The first such record
found determines the ownership.

###### BUG: if a record is linked to multiple users or groups,
ownership is undefined under the above algorithm.  If this is
an issue, the algorithm should be changed to conduct a more
exhaustive search for a matching ownership record.

=head2 user_owns() - determine whether the current user "owns" the given row

Usage: C<$db-E<gt>user_owns( TABLE, RECORD );>

TABLE is the table being operated on.  RECORD is a record id, or a
datahash reference

Returns true if:

=over 4

=item * user level 1: the record is related to the current user's member record

=item * user level 2: the record is related to the current user's group record

=item * user level 3 or 4: these users own everything

=back

This presumes a 4-level user model, which is common, but not universal.

=head2 group_owns() - determine whether the current group "owns" the given row

Usage: C<$db-E<gt>group_owns( TABLE, RECORD );>

TABLE is the table being operated on.  RECORD is a record id, or a
datahash reference

Returns true if:

=over 4

=item * user level 1 or 2: the record is related to the current user's
group record

=item * user level 3 or 4: these users own everything

=back

=head2 find_owner() - determine who owns the given row

Usage: C<$db-E<gt>find_owner( TABLE, RECORD, TYPE );>

TABLE is the table being operated on.  RECORD is a record id, or a
datahash reference.  TYPE is the ownder type, either "user" or "group".
The UID or GID of the record is returned.

=head2 find_descendants() - find all records that descend from a particular record

Usage: C<$db-E<gt>find_descendants( TABLE, RECORD );>

Each row has rows that reference it, and rows that references those,
and so on.  The find_descendants method returns pointers to all
rows that descend from a single ancestral row.  It is useful for
locating all data pertaining to a specific data item.

The return value is a reference to hash, formatted as follows:
C<$ref-E<gt>{table1} =E<gt> [ key1, key2, ... ], $ref-E<gt>{table2} =E<gt> [ ... ], ...>

=cut

#======================================================================

sub user_owns {
  my ($this,$table,$record) = @_;
  #print "checking ownership of $table $record\n";
  my $stat = $this->run_handler("user_owns",$table,$record);
  return $stat if (defined $stat);
  my $level = $this->authorize();
  my ($uid, $gid, $row_id);
  if ($level >= 3) {
    # current user is an administrator;
    # they are permitted to operate on anything
    return 1;
  }
  elsif ($level == 2) {
    # group leaders own everything in their group
    return $this->group_owns($table,$record);
  }
  elsif ($this->my_uid() == $this->find_owner($table,$record,"user")) {
    # regular user owns this row
    return 1;
  }
  else {
    # no ownership match
    return 0;
  }
}

#----------------------------------------------------------------------
# group_owns : determine whether the current user's group "owns" the
#              given row.
#
# Arguments:
# table - the table being operated on
# row - a record id, or a datahash reference
#
# Returns true if:
#   - user level 1 or 2: the record is related to the current user's
#                        group record
#   - user level 3 or 4: these users own everything

sub group_owns {
  my ($this,$table,$record) = @_;
  #print "checking ownership of $table $record\n";
  my $stat = $this->run_handler("group_owns",$table,$record);
  return $stat if (defined $stat);
  my $level = $this->authorize();
  if ($level >= 3) {
    # current user is an administrator;
    # they are permitted to operate on anything
    return 1;
  }
  if ($this->my_gid() == $this->find_owner($table,$record,"group")) {
    return 1;
  }
  else {
    return 0;
  }
}

#----------------------------------------------------------------------
# find_owner : determine who owns the given row
#
# Arguments:
# table - the table being operated on
# row - a record id, or a datahash reference
# owner_type - "user" or "group" - returns the uid or the gid that
#              owns the row, respectively.

sub find_owner {
  my ($this,$table,$record,$owner_type) = @_;
  if (!$owner_type) { $owner_type = "user"; }
  my $stat = $this->run_handler("find_owner",$table,$record,$owner_type);
  return $stat if (defined $stat);
  my $level = $this->authorize();
  my ($uid, $gid, $row_id);
  if ($owner_type eq "group") {
    $gid = $this->my_gid();
  }
  $uid = $this->my_uid();

#  my $tablemap = $this->{map}->get_table($table);

  my @parent = ($table);
  my @pid;
  if (ref $record) {
    $row_id = ($record->{$this->get_key($table)});
  }
  else {
    $row_id = ($record);
  }
  if (exists $this->{ownership}->{"$table:$row_id:$owner_type"}) {
    # already checked this row before
    return $this->{ownership}->{"$table:$row_id:$owner_type"};
  }
  @pid = ($row_id);
  my %beenthere;
  if (!exists $this->{ownership}) { $this->{ownership} = {}; }

  # perform a recursive search through all related tables
  while (@parent > 0) {
    my $parent = shift @parent;
    my $pid = shift @pid;
    #print "searching for ownership in $parent1 $pid\n";
    if (!$pid) { next; } # null reference
    if ($owner_type eq "group") {
      if ($parent eq $config{auth}->{group_table}) {
	$this->{ownership}->{"$table:$row_id:group"} = $pid;
	return $pid;
      }
    }
    else {
      if ($parent eq $config{auth}->{user_table}) {
	$this->{ownership}->{"$parent:$row_id:user"} = $pid;
	return $pid;
      }
    }
    # add parents of this record to the search list
    my @new_parents = $this->{map}->get_parents($parent);
    foreach my $new_parent (@new_parents) {
      my $col = $this->{map}->get_reference_column($parent,$new_parent);
      my %data = $this->fetch($parent,$pid);
      # watch out for infinite recursion!
      if ($beenthere{"$new_parent:$data{col}"}) { next; }
      $beenthere{"$new_parent:$data{col}"} = 1;
      push @parent,$new_parent;
      push @pid,$data{$col};
    }
  }
  # no ownership match found
  $this->{ownership}->{"$table:$row_id:$owner_type"} = 0;
  return 0;
}

#======================================================================
# DESCENDANT SEARCHES
#
# Each row has rows that reference it, and rows that references those,
# and so on.  The find_descendants method returns pointers to all
# rows that descend from a single ancestral row.  It is useful for
# locating all data pertaining to a specific data item.
#
# The return value is a reference to hash, formatted as follows:
# $ref->{table1} => [ key1, key2, ... ]
#     ->{table2} => [ ... ]
#     ...
#
# BUG: should have a flag to select exclusive/non-exclusive relationships.
# No ownership check on descendants - shouldn't matter in simple case.

sub find_descendants {
    my ($this,$table,$record) = @_;
    my $row_id;
    if (ref $record) {
	$row_id = ($record->{$this->get_key($table)});
    }
    else {
	$row_id = ($record);
    }
    my %relative;
    $relative{$table} = [ $row_id ];
    my @table_list = ($table);
    my @id_list = ($row_id);
    while (@table_list > 0) {
	my $tab = shift @table_list;
	my $key = shift @id_list;
	foreach my $child ($this->{map}->get_children($tab)) {
	    my $rlist = $this->fetch_child($child,$tab,$key);
	    my $childkey = $this->get_key($child);
	    foreach my $rdata (@$rlist) {
		# add this row to the list
		if (exists $relative{$child}) {
		    push @{$relative{$child}},$rdata->{$childkey};
		}
		else {
		    $relative{$child} = [ $rdata->{$childkey} ];
		}
		# add all children to the search list
		push @table_list,$child;
		push @id_list,$rdata->{$childkey};
	    }
	}
    }
    return \%relative;
}

#======================================================================

=pod

=head1 Sorting

The sort method orders lists of datahashes.

Usage: C<@sorted = db-E<gt>sort_rows($table,$list,@columns);>

where C<$list> is a list of datahashes, and C<@columns> is a list of columns
to sort on.  Determination of whether to use alphabetic or numeric
comparisons for a given column is done automatically, using the dbmap.

NOTE: you can also sort by passing the sortable column list to
the DB fetch/select methods.  Those sorts are performed by the DB
engine, if that is supported.  This sort method, on the other hand,
is performed by the webserver, and works for all drivers.  Which is
the more efficient approach depends on the relative loads experienced
by the two servers.

FIXME: if @columns is not defined, automatically sort on the key
display column(s)

=cut

#------------------------------------------------------------------------

sub sort_rows {
  my ($this,$table,$list,@columns) = @_;
  return () if (ref $list ne "ARRAY");
  my $desc = 0;
  for (my $c = 0; $c < @columns; $c++)
  {
	if (substr($columns[$c],0,1) eq "-")
	{
		$desc = 1;
		$columns[$c] = substr($columns[$c], 1, length($columns[$c]));
	}
  }
  # determine whether numeric comparisons are needed to sort each column
  my @numeric = ();
  foreach my $col (@columns) {
    my $datatype = $this->{map}->get($table,$col,"datatype");
    if ($datatype =~ /^(int|decimal|money|key\#)$/) { push @numeric,1; }
    else { push @numeric,0; }
  }
  return sort {
    my $stat;
    #foreach my $col (@columns) {
   for (my $i = 0; $i <= $#columns; $i++) {
      if ($numeric[$i]) {
	$stat = $desc ? $b->{$columns[$i]} <=> $a->{$columns[$i]} : $a->{$columns[$i]} <=> $b->{$columns[$i]};
      }
      else {
	$stat = $desc ? $b->{$columns[$i]} cmp $a->{$columns[$i]} : $a->{$columns[$i]} cmp $b->{$columns[$i]};
      }
      last if ($stat);
    }
    return $stat;
  } @$list;
}

#========================================================================
# Encryption Tools

#------------------------------------------------------------------------
# get_crypt_key : 
# Make the passphrase used to decode an encrypted field unique.
# If possible, hash it up a bit to obfuscate it further.
#------------------------------------------------------------------------
# REMOVED, since we can't determine the record ID on inserts.

#sub get_crypt_key {
#    my $this = shift;
#    my $key = $config{crypt_pass};
#    if ($config{crypt_unique}) {
#	# add uniqueness markers to the pass phrase
#	if (require MD5) {
#	    # use MD5 hash
#	    $key .= MD5->hexhash(join("",@_));
#	}
#	else {
#	    # use cleartext uniqueness keys
#	    $key .= join(" ",@_);
#	}
#    }
#    return $key;
#}

sub decrypt_column {
    my ($this,$data,$table,$rec,$col) = @_;
    my $access_type = $config{crypt_access};
    my $access = $this->authorize;
    # are we allowed to view encrypted data?
    if ($access_type eq "owner") {
	if ($access < $config{auth}{root_auth_level} - 1 &&
	    ! $this->user_owns($table,$rec)) {
	    $this->warn("unable to decrypt $table $rec $col");
	    return undef;
	}
    }
    else {
	if ($access < $config{auth}{root_auth_level} - 1) {
	    $this->warn("unable to decrypt $table $rec $col");
	    return undef;
	}
    }
    return $this->decrypt($data);
}

sub decrypt {
    my ($this,$data) = @_;
    if (require Crypt::CipherSaber) {
	# fetch the key to decrypt this data
	my $key = $config{crypt_pass};
	# convert the text data to binary
	my $bincrypt = decode_base64($data);
	# set encryption level
	my $level = $config{crypt_level} > 1 ? $config{crypt_level} : 1;
	# decrypt
	my $cs = Crypt::CipherSaber->new($key,$level);
	return $cs->decrypt($bincrypt);
    }
    else {
	my $error = "decryption failed: CipherSaber not found";
	$this->error($error);
	return undef;
    }
}

sub encrypt {
    my ($this,$data) = @_;
    if (require Crypt::CipherSaber) {
	my $key = $config{crypt_pass};
	my $level = $config{crypt_level} > 1 ? 
	    $config{crypt_level} : 1;
	my $cs = Crypt::CipherSaber->new($key,$level);
	my $bincrypt = $cs->encrypt($data);
	return encode_base64($bincrypt,"");
    }
    else {
	my $error = "encryption failed: CipherSaber not found";
	$this->error($error);
	# do not pass the data thru
	return undef;
    }
}

1;


