#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Misc;

use CGI;
use strict;

#========================================================================

=pod

=head1 ExSite::Misc  --  miscellaneous utility functions needed by ExSite::

Misc routines and functions that have no dependence on other ExSite code.

=cut

#========================================================================

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT);
    $VERSION     = 1.00;
    @ISA         = qw(Exporter);
    @EXPORT = qw(   &EncodeHash &DecodePost &DecodeQuery &DecodeString
		    &DecodeInput
		    &get_cookie &edit_cookie
		    &columnwise &keywise
		    &MimeType &ShowHash
		    &substitute &get_file
		    &url_escape &html_escape
		    );
}

# non-exported package globals

use vars qw($multi_sep $ncodechars @decodechars @decodecharre @encodechars
	    %substitution_method
	    );

# multiple input separator; used to join multiple data inputs from the
# same input tag name

$multi_sep = "; ";

# characters that should be encoded as %hex values
$ncodechars = 6;
@decodechars = ("%", "+", "&", "@", "\"", "\#");
@decodecharre = ("%", "\\+", "&", "@", "\"", "\#"),
@encodechars = ("%25", "%2B", "%26", "%40", "%22", "%23");

#------------------------------------------------------------------------

=pod

=head1 Hash Encoding Tools

=head2 EncodeHash(%hash) - converts a hash to a URL-encoded string

If the hash contains array references, those are converted to
comma-delimited lists of values.  If the hash contains hash
references, those are recursively converted using C<&EncodeHash>.

=head2 DecodeString($string) - converts a URL-encoded string to a hash

DecodeString can also be used to parse any string with key=value pairs
delimited by character X, if X is supplied as the 2nd argument.

Multiply-defined keys have their values appended to each other,
separated by the value of C<$multi_sep> ("; " by default).

=head2 DecodeQuery() - decode the QUERY_STRING

Passes $ENV{QUERY_STRING} through &DecodeString().

=head2 DecodePost() - decode the POST data.

For URL-encoded form data, passes the data though &DecodeString().
Multipart-encoded data is more complex (and required for file uploads);
in that case, the CGI:: package is used to parse the data.

=head2 DecodeInput() - decodes the input to the page.

Decodes either the POST or QUERY data, depending on the REQUEST_METHOD.

=cut

#------------------------------------------------------------------------

sub EncodeHash {
    # expecting a hash, or key,value pairs in input list
    my %in = @_;
    my @arglist = ();
    while ((my $key, my $val) = each %in) {
	my $arg;
	my $ref = ref($val);
	if ($ref eq "ARRAY") {
	    $arg = "$key=".join(",",@$val);
	}
	elsif ($ref eq "HASH") {
	    $arg = "$key=".&EncodeHash(%$val);
	}
	elsif ($val eq "") {
	    $arg = "$key";
	}
	else {
	    $arg = "$key=$val";
	}
	for (my $i = 0; $i < $ncodechars; $i++) {
	    $arg =~ s/$decodecharre[$i]/$encodechars[$i]/g;
	}
	$arg =~ s/ /\+/g;
	push @arglist,$arg;
    }
    return join("&",@arglist);
}

sub DecodeInput {
    # searches for input in Post first, then Query
    if ($ENV{REQUEST_METHOD} =~ /POST/i) {
        return &DecodePost;
    }
    return &DecodeQuery;
}


sub DecodePost {
    my $in;
    if ($ENV{CONTENT_TYPE} =~ /multipart\/form-data/) {
      # CONTENT_TYPE=multipart/form-data
      # (this form contains file uploads; use CGI module to decode it)
      my %data = ();
      my $in = CGI::new();
      foreach my $par ($in->param) {
	$data{$par} = $in->param($par);
      }
      return %data;
    }
    else {
      # CONTENT_TYPE=application/x-www-form-urlencoded
      my $clen = $ENV{'CONTENT_LENGTH'};
      if ($clen) {
	read(STDIN,$in,$ENV{'CONTENT_LENGTH'});
	return &DecodeString($in);
      }
    }
    return ();
}

sub DecodeQuery {
    my $query = $ENV{QUERY_STRING};
    if (defined $query) {
	return &DecodeString($query);
    }
    return ();
}

sub DecodeString {
    my $sep = $_[1] || "&";
    my @in = split(/$sep/,$_[0]);
    my %in;
    foreach my $i (0 .. $#in) {
        # Convert pluses to spaces
        $in[$i] =~ s/\+/ /g;

        # Split into key and value.
        (my $key, my $val) = split(/=/,$in[$i],2); # splits on the first =

        # Convert %XX from hex numbers to alphanumeric
        $key =~ s/%(..)/pack("c",hex($1))/ge;
        $val =~ s/%(..)/pack("c",hex($1))/ge;

        # Associate key and value. \0 is the multiple separator
        $in{$key} .= $multi_sep if (defined($in{$key}));
        $in{$key} .= $val;
    }
    return %in;
}

#------------------------------------------------------------------------

=pod

=head1 Datahash Array Conversion Functions:

=head2 columnwise($ref) - convert a list of datahashes into a hash of datalists

input: an array of hash references with identical keys.

output: a hash of arrays, indexed by the same keys.  For instance, 
C<( { a=E<gt>1, b=E<gt>2 }, { a=E<gt>3, b=E<gt>4 }, { a=E<gt>5, b=E<gt>6 } )>
becomes  C<( a =E<gt> [1,3,5], b =E<gt> [2,4,6] )>

=head2 keywise($key,$ref) - convert a list of datahashes into a hash of datahashes

input: a hash key to index on, and an array of hash references

output: a hash of hashes, indexed by one of the keys.  It is assumed
        that the value associated with this key is unique across all
        the hashes.  If not, the last such key is used.

For example, taking C<( { a=E<gt>1, b=E<gt>2 }, { a=E<gt>3, b=E<gt>4}, { a=E<gt>5, b=E<gt>6 } )> 
and indexing it by key "a", gives 
C<{ 1=E<gt>{a=E<gt>1,b=E<gt>2}, 3=E<gt>{a=E<gt>3,b=E<gt>4 }>.

=cut

#------------------------------------------------------------------------

sub columnwise ($) {
    my ($ref_list_datahash) = shift;
    my @keys = keys %{$ref_list_datahash->[0]};
    my %datahash;
    foreach my $key (@keys) {
      $datahash{$key} = [];
      for (my $i = 0; $i < @$ref_list_datahash; $i++) {
	push @{$datahash{$key}}, $ref_list_datahash->[$i]->{$key};
      }
    }
    return \%datahash;
}

sub keywise ($$) {
    my ($key,$ref_list_datahash) = @_;
    my %datahash = ();
    foreach my $rhash (@{$ref_list_datahash}) {
      $datahash{$rhash->{$key}} = $rhash;
    }
    return \%datahash;
}

#------------------------------------------------------------------------

=pod

=head1 Miscellaneous other functions...

=head2 get_cookie($name,$sepchar) - retrieve a cookie value

Returns the value of cookie "$name".
If sepchar is defined, the cookie value is split on that character,
and an array of values is returned.

=head2 edit_cookie($name,$value) - changes current cookie value

This changes the value of cookie $name to $value.  It ONLY changes
the value for the current request - it does not re-set the coookie for 
subsequent requests.  To modify the cookie for future requests, you
must also call &ExSite::Config::set_cookie(). &edit_cookie() returns 
the new cookie value, which can be passed to set_cookie, if needed.

=head2 MimeType($filename) - guess the MIME type of a file

This routine uses the filename (in particular, the suffix) to guess
the MIME type of the file.

=head2 ShowHash() - display a hash's contents in HTML format

A hash or hash reference may be passed to this routine;  a block of
HTML is returned.

=cut

#------------------------------------------------------------------------

sub get_cookie { 
    my ($name,$sepchar) = @_;
    if (exists $ENV{'HTTP_COOKIE'}) {
        my @cookies = split /; /, $ENV{'HTTP_COOKIE'};
        foreach my $cookie (@cookies) {
            if ($cookie =~ /^$name=/) {
		my (undef,$value) = split /=/,$cookie,2;
		if ($sepchar) { 
		    return split /$sepchar/,$value;
		}
		else {
		    return $value;
		}
            }
        }
    }
    return undef;
}

sub edit_cookie { 
    my ($name,$value) = @_;
    my @cookies;
    my $done;
    if (exists $ENV{'HTTP_COOKIE'}) {
        @cookies = split /; /, $ENV{'HTTP_COOKIE'};
        foreach my $cookie (@cookies) {
            if ($cookie =~ /^$name=/) {
		$cookie = "$name=$value";
		$done = 1;
		last;
	    }
        }
    }
    if (! $done) {
	# add the cookie
	push @cookies, "$name=$value";
    }
    # redefine the cookie
    my $new_cookie = join "; ",@cookies;
    $ENV{HTTP_COOKIE} = $new_cookie;
    return $new_cookie;
}

sub MimeType ($) {
  $_ = shift;
  if    (/\.txt$/i)   { return "text/plain"; }
  elsif (/\.html?$/i) { return "text/html"; }
  elsif (/\.css$/i)   { return "text/css"; }
  elsif (/\.tex$/i)   { return "application/x-latex"; }
  elsif (/\.doc$/i)   { return "application/msword"; }
  elsif (/\.ppt$/i)   { return "application/powerpoint"; }
  elsif (/\.pdf$/i)   { return "application/pdf"; }
  elsif (/\.e?ps$/i)  { return "application/postscript"; }
  elsif (/\.dvi$/i)   { return "application/x-dvi"; }
  elsif (/\.gif$/i)   { return "image/gif"; }
  elsif (/\.jpe?g$/i) { return "image/jpeg"; }
  elsif (/\.png$/i)   { return "image/png"; }
  elsif (/\.mpe?g?$/i){ return "video/mpeg"; }
  elsif (/\.mov$/i)   { return "video/quicktime"; }
  elsif (/\.qt$/i)    { return "video/quicktime"; }
  elsif (/\.avi$/i)   { return "video/x-msvideo"; }
  elsif (/\.au$/i)    { return "audio/basic"; }
  elsif (/\.snd$/i)   { return "audio/basic"; }
  elsif (/\.aif.?$/i) { return "audio/x-aiff"; }
  elsif (/\.wav$/i)   { return "audio/x-wav"; }
  else                { return "application/octet-stream"; }
}

# return the standard filename suffix used by a mime type

sub FileToMime ($) {
  $_ = shift;
  if    (/text\/plain/i)             { return "txt"; }
  elsif (/text\/html/i)              { return "html"; }
  elsif (/text\/css/i)               { return "css"; }
  elsif (/application\/x-latex/i)    { return "tex"; }
  elsif (/application\/msword/i)     { return "doc"; }
  elsif (/application\/powerpoint/i)  { return "ppt"; }
  elsif (/application\/pdf/i)         { return "pdf"; }
  elsif (/application\/postscript/i)  { return "ps"; }
  elsif (/application\/x-dvi/i)       { return "dvi"; }
  elsif (/image\/gif/i)               { return "gif"; }
  elsif (/image\/jpeg/i)              { return "jpg"; }
  elsif (/image\/png/i)               { return "png"; }
  elsif (/video\/mpeg/i)              { return "mpeg"; }
  elsif (/video\/quicktime/i)         { return "mov"; }
  elsif (/video\/x-msvideo/i)         { return "avi"; }
  elsif (/audio\/basic/i)             { return "au"; }
  elsif (/audio\/x-aiff/i)            { return "aif"; }
  elsif (/audio\/x-wav/i)             { return "wav"; }
  else                               { return "dat"; }
}

sub ShowHash {
    my $hash;
    if ($#_ > 0) { my %h = @_; $hash = \%h; }
    else { $hash = shift; }
    if (ref $hash !~ /HASH/) {
	return "Not a hash reference: $hash";
    }
    my $out = "<UL>\n";
    foreach my $key (keys %{$hash}) {
	$out .= "<LI><B>&rarr;{".$key."}</B>";
	if ($hash->{$key} =~ /HASH/) {
	    $out .= &ShowHash($hash->{$key});
	} elsif ($hash->{$key} =~ /ARRAY/) {
	    my $data = " = [ARRAY](".join(', ',@{$hash->{$key}}).")";
	    $data =~ s/</&lt;/g;
	    $data =~ s/>/&gt;/g;
	    $out .= $data;
	} else {
	    my $data = " = \"$hash->{$key}\"";
	    $data =~ s/</&lt;/g;
	    $data =~ s/>/&gt;/g;
	    $out .= $data;
	}
	$out .= "</LI>\n";
    }
    $out .= "</UL>\n";
    return $out;
}

#------------------------------------------------------------------------
# get_file : return the contents of a file, modified according to the
#            calling options:
#
#------------------------------------------------------------------------

sub get_file($$) {
    my ($file, $opt) = @_;
    my $len = -s $file;
    open F,"<$file" or return;
    my $data;
    read(F,$data,$len) or return;
    close F or return;
    if ($opt->{imgpath}) {
        $data =~ s/(SRC=\"?)/$1$opt->{imgpath}\//g;
    }
    return $data;
}

#------------------------------------------------------------------------
# url_escape : escape characters using URL escape sequences
#
# $text = string to escape
# $echars = contains characters to escape [ %+&@"] by default
#------------------------------------------------------------------------

sub url_escape($$) {
    my ($text,$echars) = @_;
    if (! $echars) { $echars = '%+ &@"'; }
    my $rechars = '[]+*^$';
    for (my $i = 0; $i < length($echars); $i++) {
	my $char = substr($echars,$i,1);
	#my $esc = ($char eq " " ? '+' : sprintf("%%%02x",ord $char));
	my $esc = sprintf("%%%02x",ord $char);
	if (index($rechars,$char) > -1) { $char = "\\$char"; }
	$text =~ s/$char/$esc/g;
    }
    return $text;
}

#------------------------------------------------------------------------
# html_escape : escape characters using HTML escape sequences
#
# $text = string to escape
# $echars = contains characters to escape [&<>] by default
#------------------------------------------------------------------------

sub html_escape($$) {
    my ($text,$echars) = @_;
    if (! $echars) { $echars = '&<>'; }
    my $rechars = '[]+*^$';
    for (my $i = 0; $i < length($echars); $i++) {
	my $char = substr($echars,$i,1);
	my $esc = sprintf("&%d;",ord $char);
	if (index($rechars,$char) > -1) { $char = "\\$char"; }
	$text =~ s/$char/$esc/g;
    }
    return $text;
}

#------------------------------------------------------------------------
# substitute : replace markers with values in a string
#
# Replaces marker tag [[tag]] with the value in $subhash{tag}.
# Does so for all such tags in the text.  Tags may contain word 
# characters only (alphanumeric plus _).
#
# $text = string to perform substitutions on
# $subhash = hash of marker keys=>values
# $startmark = substring denoting the start of a marker tag ("[[" default)
# $endmark = substring denoting the end of a marker tag ("]]" default)
#------------------------------------------------------------------------

sub substitute {
    my ($text,$subhash,$startmark,$endmark) = @_;
    if (! $startmark) { $startmark = '\[\['; }
    if (! $endmark)   { $endmark = '\]\]'; }
    $text =~ s/$startmark(\w+)$endmark/$subhash->{$1}/g;
    return $text;
}

1;


