#------------------------------------------------------------------------
# ExSite::SQL
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
#
# Generic SQL Driver for ExSite web development package.
# (Designed to work with MySQL, but should be easily adaptable to any
# SQL engine, since it goes through Perl DBI.)
#
# ExSite DB drivers have the following methods:
#
# new : make a new database handle
# handler : handler installation method
# get_tables : retrieve all table names
# get_columns : retrieve all column names from a table
# get_key : retrieve the primary key for a table
# select : retrieve data from tables
# insert : add new data to tables
# update : modify old data in tables
# delete : remove data from tables
#
# See the ExSite::DB package for additional wrapper and convenience
# methods to fetch data from and modify the database.
#
# SECURITY NOTE: since the DB package is unaware of the underlying
# engine, it cannot perform proper security checks on the data being
# passed to the DB server.  Therefore all SQL parameters must be 
# validated within each driver, ie. in this file.
#------------------------------------------------------------------------

package ExSite::SQL;

use strict;
use Carp;
use DBI;
use ExSite::Config;
use ExSite::Base;
our @ISA = qw(ExSite::Base);

#------------------------------------------------------------------------
# new : make a new SQL driver handle
#
# all configuration parameters are passed in a hash.  A typical
# invocation might look like this:
#
# (name=>"my_db",server=>"my.server.com",user=>"my_name",pass=>"my_pwd")
#------------------------------------------------------------------------

sub new($%) {
  my($this, %db) = @_;
  my($obj,$class);

  $obj = {};
  $class = ref($this) || $this;
  bless $obj, $class;
 $db{debug}=1;
  if ($db{debug}) {
    $obj->{debug} = $db{debug};
    delete $db{debug};
    #$obj->trace(2);
  }
  # store most-recent operation status
  $obj->{query_error} = undef;
  # table info cache
  $obj->{sql_table_info} = {};

  # Connect!
  # typical $connect values
  # dbi:DriverName:database_name
  # dbi:DriverName:database_name@hostname:port
  # dbi:DriverName:database=database_name;host=hostname;port=port
  # BUT . . .
  # I'm following MySQL format by default and inserting
  #     modifications for other drivers
  # dbi:DriverName:database_name;host=hostname;port=port

  # default to MySQL
  $db{'driver'} = $db{'driver'} ? $db{'driver'} : "mysql";

  my $connect = join(":", "dbi", $db{'driver'}, $db{'name'});
  if($db{'server'}) {
    $connect .= ";host=".$db{'server'};
    if($db{'port'}) {
      $connect .= ";".$db{'port'};
    }
  }
  $obj->{dbh} = DBI->connect($connect, $db{'user'}, $db{'pass'},
			     { AutoCommit=>1, RaiseError=>0, PrintError=>0 });
  if (! $obj->{dbh}) {
    $obj->error("ExSite::SQL - failed to connect: $DBI::errstr; $!");
  }
  $obj->info("database connection established");

  # store db access parameters for later lookup
  $obj->{db} = {%db};

  return $obj;
}

#------------------------------------------------------------------------
# query : generic query method
#
# Prepares and executes a SQL statement.  Status and errors are stored
# in the object for later reporting.  Returns a statement handle.
#
# This method is intended for internal use, but it could be called
# externally by anyone familiar with DBI methods.
#------------------------------------------------------------------------

sub query($$) {
  my ($this, $sql) = @_;
  my ($sth);
  # clear the error status
  $this->{query_error} = undef;
  # prepare the statement
  $sth = $this->{dbh}->prepare($sql);
  #$this->info($sql);
  if (! $sth) {
    # failure
    $this->{query_error} = "ExSite::SQL::query - failed to prepare \"$sql\"; $DBI::errstr; $!";
    $this->error($this->{query_error});
  }
  # execute the statement
  if (! $sth->execute) {
    # failure
    $this->{query_error} = "ExSite::SQL::query - failed to execute \"$sql\"; $DBI::errstr; $!";
    $this->error($this->{query_error});
  }
  return $sth;
  # calling routine is responsible for getting the results, if any
}

#------------------------------------------------------------------------
# get_tables : return list of tables in the database
#------------------------------------------------------------------------

sub get_tables($) {
  my ($this) = shift;
  my $sth = $this->query("SHOW TABLES");
  return () if ($this->{query_error});
  my @tables = ();
  # return value has ridiculous formatting; extract important info
  foreach my $tab (@{$sth->fetchall_arrayref}) {
    push @tables, $tab->[0];
  }
  return @tables;
}

#------------------------------------------------------------------------
# get_table_info : fetch general information about the columns in a table
# Used internally by get_key and get_columns methods.
#------------------------------------------------------------------------

sub get_table_info($$) {
  my ($this, $table) = @_;
  if ($table !~ /^\w+$/) {
      # validate args
      $this->error("SQL::get_columns: invalid table: $table");
      return;
  }
  if (! $this->{sql_table_info}->{$table}) {
    my $sth = $this->query("SHOW COLUMNS FROM $table");
    return if ($this->{query_error});
    $this->{sql_table_info}->{$table} = [];
    while (my $row = $sth->fetchrow_hashref) {
      push @{$this->{sql_table_info}->{$table}},$row;
    }
  }
}

#------------------------------------------------------------------------
# get_columns : return list of columns in a table
#------------------------------------------------------------------------

sub dbget_columns($$) {
  my ($this, $table) = @_;
  if ($table !~ /^\w+$/) {
      # validate args
      $this->error("SQL::get_columns: invalid table: $table");
      return;
  }
  $this->get_table_info($table);
  return if ($this->{query_error});
  my @columns = ();
  foreach my $col (@{$this->{sql_table_info}->{$table}}) {
    push @columns, $col->{Field};
  }
  return @columns;
}

#------------------------------------------------------------------------
# get_key : return primary key of a table
#------------------------------------------------------------------------

sub dbget_key($$) {
  my ($this, $table) = @_;
  if ($table !~ /^\w+$/) {
      # validate args
      $this->error("SQL::get_key: invalid table: $table");
      return;
  }
  $this->get_table_info($table);
  return undef if ($this->{query_error});
  foreach my $col (@{$this->{sql_table_info}->{$table}}) {
    # assume key info is stored in column "Key"
    if ($col->{Key} =~ /^pri/i) {
      # this appears to be a primary key
      return $col->{Field};
    }
  }
  return undef;
}

#-------------------------------------------------------------------------
# where_match : build a WHERE clause from a match hash
#
# parameters:
# $match = reference to a match hash
#
# Used internally by the select, update, and delete methods.
#-------------------------------------------------------------------------

sub where_match ($$) {
  my ($this,$match) = @_;
  my @wlist = ();
  my $where;
  while (my ($key, $val) = each(%$match)) {
    # sanity check on keys
    if ($key !~ /^[\w\.]+$/) {
	$this->error("SQL::where_match: invalid column: $key");
	return undef;
    }	
    # default operator
    my $op = "=";
    # if wildcards are present, use "LIKE"
    if($val =~ /%/) {
      if($val =~ /^%|[^\\]%/) {
	# unescaped wildcard
	$op = "LIKE";
	# quote while protecting "\%"
	# (quote() escapes the backslash otherwise)
	my $ph = "~_jwosja_~"; # a highly improbable string unaffected by quote()
	$val =~ s/\\%/$ph/g;
	$val = $this->{dbh}->quote($val);
	$val =~ s/$ph/\\%/g;
      }
      else {
	# strip backslashes before quoting; "%" is not a meta-char for "="
	$val =~ s/\\%/%/g;
	$val = $this->{dbh}->quote($val);
      }
    }
    elsif (ref $val) {
      # if val is a reference, it points to a SQL literal
      # WARNING! no sanity check
      $val = $$val;
    }
    else {
      $val = $this->{dbh}->quote($val);
    }
    # build the where
    push(@wlist, "$key $op $val");
  }
  if (scalar @wlist) {
    $where = "WHERE ".join(" && ", @wlist);
  }
  return $where;
}

#-------------------------------------------------------------------------
# multi-purpose select function
#
# parameters:
# $rtable = table, or reference to array of tables
# $rcolumn = column, or reference to array of columns, or reference to a
#            hash of table=>columnlist, or NULL for all columns
# $cond = explicit conditional, or a reference to a match hash, which
#            can optionally contain wildcards
# $sort = explicit sort string, or a reference to a list of columns
# $limit = maximum number of rows to fetch
#
# Only $rtable is required for the function to work.
#
# This gives you a fairly flexible set of select and join options,
# without being completely generic.
#-------------------------------------------------------------------------

sub dbselect($$;$$$$) {
    my ($this,$rtable,$rcolumn,$cond,$sort,$limit) = @_;

    my ($tlist,$clist,$where,$order,$sql,@result);
    # table(s) to select from
    if (ref $rtable eq "ARRAY") {
      # array of tables was passed
      $tlist = join(",",@$rtable);
      if ($tlist !~ /^(\w+,)*\w+$/) {
	  # validate 
	  $this->error("SQL::select: invalid table list: $tlist");
	  return undef;
      }
    }
    else {
      # single table was passed as a scalar
      $tlist = $rtable;
      if ($tlist !~ /^\w+$/) {
	  # validate 
	  $this->error("SQL::select: invalid table: $tlist");
	  return undef;
      }
    }

    # column(s) to select
    if (ref $rcolumn eq "ARRAY") {
      # array of columns was passed
      $clist = join(",",@$rcolumn);
      if ($clist !~ /^([\w\.]+,)*[\w\.]+$/) {
	  $this->error("SQL::select: invalid column list: $clist");
	  return undef;
      }
    }
    elsif (ref $rcolumn eq "HASH") {
      # hash of table => reference_to_column_list was passed
      my @joinlist = ();
      while (my ($table,$list) = each %$rcolumn) {
	  if ($table !~ /^\w+$/) {
	      $this->error("SQL::select: invalid table: $table");
	      return undef;
	  }
	  foreach my $col (@$list) {
	      if ($col !~ /^\w+$/) {
		  $this->error("SQL::select: invalid column: $col");
		  return undef;
	      }	
	      push @joinlist,"$table.$col";
	  }
      }
      $clist = join(",",@joinlist);
    }
    elsif ($rcolumn) {
	# single column or explicit function was passed as a scalar
	if ($rcolumn !~ /^\w+$/) {
	    ####### this regexp will not accept functions!
	    $this->error("SQL::select: invalid column: $rcolumn");
	    return undef;
	}
	$clist = $rcolumn;
    }
    else {
      # no columns passed, select all columns
      $clist = "*";
    }

    # condition to select on
    if (ref $cond eq "HASH") {
      # condition was passed as a match hash -
      # the selected records must match all elements of the hash.
      $where = $this->where_match($cond);
      if (! $where) {
	  $this->error("SQL::select: invalid where clause");
	  return undef;
      }	
    }
    elsif ($cond) {
      # assume condition is a scalar explicitly stating the condition.
      # WARNING: use of this feature is potentially non-portable to other
      # ExSite DB drivers.
      # Also, no sanity check on $cond, so be careful about using tainted data
      $where = "WHERE $cond";
    }

    # sort, if asked
    if ($sort) {
      my @ords = ();
      if (ref $sort eq "HASH") {
	# sort is a hash of column => direction
	while (my ($col, $dir) = each(%$sort)) {
	  if ($col !~ /^\w+$/ || $dir !~ /^(asc|desc)$/i) {
	      $this->error("SQL::select: invalid sort directive: $col $dir");
	      return undef;
	  }
	  push @ords, "$col $dir";
        }
	$order = join(",", @ords);
      }
      elsif (ref $sort eq "ARRAY") {
	# sort is a list of columns to sort in default order
	foreach my $col (@$sort) {
	    if ($col =~ /^\w+$/) {
		push @ords, $col;
	    }
	    elsif (ref($col) eq "HASH") {
	    	while (my($key,$val) = each(%$col)) {
		    if ($key !~ /^\w+$/ || $val !~ /^(asc|desc)$/i) {
		 	$this->error("SQL::select: invalid sort column: $col");
			return undef;
		    }
		    push @ords, "$key $val";
		}
	    }
	    else{
		$this->error("SQL::select: invalid sort column: $col");
		return undef;
	    }
        }
	$order = join(",", @ords);
      }
      else {
	# explicit sort list
	# WARNING: may not be portable to other ExSite DB drivers.
	if ($sort !~ /^(\w+( (asc|desc))?,\s*)*(\w+( (asc|desc))?\s*)$/i) {
	    $this->error("SQL::select: invalid sort directive: $sort");
	    return undef;
	}
	$order = $sort;
      }
      $order = "ORDER BY $order";
    }
	
    # limit, if asked
    if ($limit) {
	if ($limit !~ /^(\d+,\s*)\d+$/) {
	    $this->error("SQL::select: invalid limit directive: $limit");
	    return undef;
	}
	$limit = "LIMIT $limit";
    }

    # execute the query
    if ($this->{handler}->{pre_select}) {
      &{$this->{handler}->{pre_select}}(@_);
    }
    $sql = "SELECT $clist FROM $tlist $where $order $limit";
    # log the query
    $this->info($sql);
    my $sth = $this->query($sql);
    @result = ();
    my $row;
    while ($row = $sth->fetchrow_hashref) {
      push @result, $row;
    }
    if ($this->{handler}->{post_select}) {
      &{$this->{handler}->{post_select}}(@_);
    }
    if (wantarray) { return @result; }
    return \@result;
}

#-------------------------------------------------------------------------
# generic insert function
#
# parameters:
# $table = a table name
# $data = reference to a data hash
#
# returns primary key of inserted record.
#-------------------------------------------------------------------------

sub dbinsert($$$) {
  my ($this,$table,$data) = @_;
  if (ref $data eq "HASH") {
    # pre-handler?
    if ($this->{handler}->{pre_insert}) {
      my $ind = &{$this->{handler}->{pre_insert}}(@_);
    }
    my $clist = join(",", keys %$data);
    if ($clist !~ /^(\w+,)*\w+$/) {
	$this->error("SQL::insert: invalid column list: $clist");
	return 0;
    }
    # values
    my @values = ();
    for my $val (values %$data) {
	push @values, $this->{dbh}->quote($val);
    }
    my $vlist = join(",", @values);

    # if primary key is defined, use "REPLACE" instead of "INSERT"
    my $pkey = $this->get_key($table);
    my $operation;
    if ($data->{$pkey}) {
      $operation = "REPLACE";
    }
    else {
      $operation = "INSERT";
    }

    my $sql = "$operation INTO $table ($clist) VALUES ($vlist)";
    
#	print "content-type:text/html\n\n" if ($this->{debug});
#    print "$sql<p>\n" if ($this->{debug});

    # Lock
    $this->query("LOCK TABLES $table WRITE");
    # Insert
    #print "Content-type: text/plain\n\n$sql\n";
    #my $new_id = 667;
    $this->query($sql);
    my $new_id = $this->{dbh}->{mysql_insertid};
    $this->{last_insert_id} = $new_id;
    # Unlock
    $this->query("UNLOCK TABLES");
    # log the insert
    $this->info($sql);
    # post-handler?
    if ($this->{handler}->{post_insert}) {
      &{$this->{handler}->{post_insert}}(@_);
    }
    return $new_id;
  }
  else {
    $this->error("ExSite::SQL::insert: bad data reference");
    return 0;
  }
}

#-------------------------------------------------------------------------
# generic update function
#
# parameters:
# $table = a table name
# $data = reference to a data hash
# $cond = explicit condition, or reference to a match hash, or NULL to
#         update the record whose primary key matches the data
#-------------------------------------------------------------------------

sub dbupdate($$$;$) {
  my ($this,$table,$data,$cond) = @_;
  my ($set,$where);
  # data
  if (ref $data eq "HASH") {
    my @set = ();
    while (my ($key,$val) = each %$data) {
      if ($key !~ /^\w+$/) {
	my $error = "SQL::update: invalid column: $key";
	$this->error($error);
	return $error;
      }
      my $qval = $this->{dbh}->quote($val);
      push @set, "$key=$qval";
    }
    $set = join(",",@set);
  }
  else {
    my $error  = "ExSite::SQL::update: bad data reference";
    $this->error($error);
    return $error;
  }

  # condition
  if (ref $cond eq "HASH") {
    # condition is a match hash
    $where = $this->where_match($cond);
    if (! $where) {
	return "SQL::update: aborted due to bad where condition";
    }
  }
  elsif ($cond) {
    # condition is an explicit WHERE clause
    #### danger: no sanity check!
    $where = $cond;
  }
  else {
    # no condition!  Select rows matching the primary key.
    my $pkey = $this->get_key($table);
    if (exists $data->{$pkey}) {
      $where = "WHERE $pkey=".$this->{dbh}->quote($data->{$pkey});
    }
    else {
      # no primary key!  Select all rows (!!!)
      #$where = "";
      #$this->warn("ExSite::SQL::update: warning! unconditional update!");
      # if you'd rather this be an error, replace with this:
      my $error = "ExSite::SQL::update: trapped unconditional update";
      $this->error($error);
      return $error;
    }
  }

  my $sql = "UPDATE $table SET $set $where";

  # pre-handler?
  if ($this->{handler}->{pre_update}) {
    &{$this->{handler}->{pre_update}}(@_);
  }
  # Lock
  $this->query("LOCK TABLES $table WRITE");
  # update
  $this->query($sql);
  my $stat = $this->{query_error};
  # Unlock
  $this->query("UNLOCK TABLES");
  # log the update
  $this->info($sql);
  # post-handler?
  if ($this->{handler}->{post_update}) {
    &{$this->{handler}->{post_update}}(@_);
  }
  return $stat;
}

#-------------------------------------------------------------------------
# generic delete function
#
# parameters:
# $table = a table name
# $cond = explicit condition, or reference to a match hash
#-------------------------------------------------------------------------

sub dbdelete($$$;$) {
  my ($this,$table,$cond) = @_;
  my $where;

  # condition
  if (ref $cond eq "HASH") {
    # condition is a match hash
    $where = $this->where_match($cond);
    if (! $where) {
	my $error = "SQL::delete: invalid delete condition";
	$this->error($error);
	return $error;
    }
  }
  elsif ($cond) {
    # condition is an explicit WHERE clause
    ##### no sanity check!
    $where = $cond;
  }
  else {
    #$this->warn("ExSite::SQL::update: warning! unconditional delete!");
    my $error = "ExSite::SQL::delete: trapped unconditional delete";
    $this->error($error);
    return $error;
  }

  my $sql = "DELETE FROM $table $where";

  # pre-handler?
  my $stat;
  if ($stat = $this->run_handler("pre_delete",$table,$cond)) {
      return $stat;
  }
  # Lock
  $this->query("LOCK TABLES $table WRITE");
  # Insert
  $this->query($sql);
  $stat = $this->{query_error};
  # Unlock
  $this->query("UNLOCK TABLES");
  # log the query
  $this->info($sql);
  # post-handler?
  $stat .= $this->run_handler("post_delete",$table,$cond);
  return $stat;
}

#-------------------------------------------------------------------------
# emergency fallback - explicit SQL query 
# (NOT PORTABLE TO OTHER EXSITE DRIVERS)
#
# DO YOUR OWN TAINT CHECKING
#-------------------------------------------------------------------------

sub custom_query($$){
  my ($this,$string)=@_;
  my ($sth, $rowvP, @rowvPs);
  $sth = $this->query($string);
  # log the query
  $this->info($string);
  ### return the data as a reference to an array of hash references
  while($rowvP = $sth->fetchrow_hashref) { push @rowvPs, {%$rowvP};}
  $sth->finish;
  if (wantarray) { return @rowvPs; }
  return \@rowvPs;
}

1;
