#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Section;

# new, setup, setupSection (preload index page)
# filename - name of directory
# httppath - htdocs path to section
# diskpath - full disk path
# get_url - full web path
# publish - write out all files

########### needed:
# get_webmasters
# is_webmaster
# section_path
# get_pages ! this section only
# get_pages_from_path ! all sections in path

use strict;
use ExSite::Config;
#use ExSite::Misc;
use ExSite::Page;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::Page);

# new() constructor is inherited from ContentData class.

sub setup {
    # so that we can do setup using a non-overloaded method
    &setupSection(@_);
}

sub setupSection { 
    my ($this,%option) = @_;

    my $section;
    if ($option{id}) {
	# lookup section record
	$section = $this->fetch("section",$option{id});
    }
    elsif (ref $option{section} eq "HASH") {
	# section record was passed
	if (ref $option{section} eq "HASH") {
	    $section = $option{section};
	}
	else {
	    $this->error("invalid section parameters passed to setupSection");
	}
    }
    if ($section) {
	$this->info("setting up section $section->{section_id}");
	$this->{data} = $section;
	$this->set_context("none");
    }
    else {
	$this->error("section $section->{section_id} not found");
    }
}

# previous - change the section object to next one up in path

sub previous {
    my $this = shift;
    my $section = $this->get;
    if ($section->{parent_id}) {
	$this->setupSection(id=>$section->{parent_id});
	return $section->{parent_id};
    }
    else {
	return undef;
    }
}

# name of section directory - overloaded variation on ContentData method

sub filename {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	if ($section->{type} =~ /native/) {
	    return $section->{publish_to};
	}
    }
    # filename is not defined for other website types
    return undef;
}

# disk path to section

sub diskpath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	# no disk path for templates and remote sites
	return undef if ($section->{type} =~ /template|remote/);
	if ($section->{type} =~ /stand-alone/) {
	    # stand-alone sites hold full path
	    return $section->{publish_to};
	}
	else {
	    if ($section->{parent_id}) {
		# this section nests within the parent section
		my $psec = new ExSite::Section(id=>$section->{parent_id});
		my $parent_path = $psec->diskpath;
		# if parent is undefined, that's because we are nested 
		# under a non-local website
		return undef if ! defined $parent_path;
		return "$parent_path/$section->{publish_to}";
	    }
	    elsif ($section->{publish_to}) {
		return "$config{server}{HTMLroot}$config{server}{HTMLpath}/$section->{publish_to}";
	    }
	    else {
		# root section of the native site
		return "$config{server}{HTMLroot}$config{server}{HTMLpath}";
	    }
	}
    }
    return undef;
}

# web path to section   ######### DEPRECATED ###########

sub httppath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	# can only give a partially-qualified URL for native sites
	if ($section->{type} =~ /native/) {
	    if ($section->{parent_id}) {
		# this section nests within the parent section
		my $psec = new ExSite::Section(id=>$section->{parent_id},
					       context=>"none");
		my $parent_path = $psec->httppath;
		# if parent is undefined, that's because we are nested 
		# under a non-native website
		return undef if ! defined $parent_path;
		return "$parent_path/$section->{publish_to}";
	    }
	    elsif ($section->{publish_to}) {
		return "$config{server}{HTMLpath}/$section->{publish_to}";
	    }
	    else {
		# root section of the native site - return base HTML path
		return $config{server}{HTMLpath};
	    }
	}
	elsif ($section->{type} !~ /template/) {
	    return $section->{url};
	}
    }
    return undef;
}

# overloaded get_url: return a URL for the section

sub get_url {
    my $this = shift;
    my $url;
    my $section = $this->get;
    return undef if ($section->{type} =~ /template/);
    while ($section) {
	if ($section->{type} =~ /native/) {
	    if ($section->{publish_to}) {
		$url = "/".$section->{publish_to}.$url;
	    }
	}
	else {
	    $url = $section->{url}.$url;
	    last;
	}
	if ($section->{parent_id}) {
	    $section = $this->fetch("section",$section->{parent_id});
	}
	else { 
	    $section = undef;
	}
    }
    if (! $url || $url =~ /^\/\w/) {
	#$url = $config{server}{HTMLpath}.$url;
	# DANGER: admin server may not be the same as public server
	$url = $config{server}{server}.$config{server}{HTMLpath}.$url;
    }
    elsif ($url =~ /^\w+\./) {
	$url = "http://$url";
    }
    return $url;
}

# sectionpath : return an array of the sites/sections traversed to get to 
#               this site

sub sectionpath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	my @parent_path;
	if ($section->{parent_id}) {
	    my $psec = new ExSite::Section(id=>$section->{parent_id});
	    @parent_path = $psec->sectionpath;
	}
	return (@parent_path,$section);
    }
}

# fetch_library_content : grabs all content objects in all libraries in the
#                         site path.  This could be useful for inserting 
#                         shared content into pages, or it could be a pig.

sub fetch_library_content {
    my ($this,%match) = @_;
    my @sections = $this->sectionpath;
    my @libs;
    foreach my $s (@sections) {
	push @libs, $this->fetch_match("page",{type=>"library",section_id=>$s->{section_id}});
    }
    my @Content;
    foreach my $l (@libs) {
	my @cont = $this->fetch_match("content",{page_id=>$l->{page_id},%match});
	foreach my $c (@cont) {
	    push @Content, new ExSite::Content(content=>$c);
	}
    }
    return @Content;
}

# get all pages for the section - returns array of Page objects
#
# %opt is a match hash for pages

sub get_all_pages {
    my ($this,%opt) = @_;
    if (! $opt{section_id}) {
	$opt{section_id} = $this->id;
    }
    my @pages = $this->fetch_match("page",\%opt);
    my @Page;
    foreach my $page (@pages) {
	push @Page, new ExSite::Page(page=>$page);
    }
    return @Page;
}

# write the section to disk

sub publish {
    my ($this,$opt) = @_;
    my $out;
    my $section = $this->get;
    if (! $section) { 
	$this->error("Section::publish: no section to publish");
	print "No section to publish!";
	return;
    }
    elsif (! $section->{type} =~ /template/) { 
	$this->error("Section::publish: cannot publish website templates");
	print "Cannot publish website templates!";
	return;
    }
    else { # if native website
	my $diskpath = $this->diskpath;
	if (! $diskpath) { 
	    $this->error("Section::publish: non-local publishing not supported");
	    print "Non-local publishing not supported yet.";
	    return;
	}
	$this->info("Section::publish: publishing section $section->{section_id}");
	# 1) create the section directories
	my $mkdir = $this->mkdir($diskpath);
	if (! $mkdir) {
	    print $this->show_diagnostics("error","html");
	    return;
	}
	print $mkdir;
	return if ($opt eq "nofiles");
	# 2) publish all pages 
	# autoflush stdout so we can monitor the progress...
	my $old_flush = $|;
	$| = 1;
	# first, libraries
	my @Page = $this->get_all_pages(type=>"library",
					status=>"active"); ### access=>"public");
	foreach my $p (@Page) {
	    print $p->publish;
	}
	# next, templates
	my @Page = $this->get_all_pages(type=>"template",
					status=>"active"); ### access=>"public");
	foreach my $p (@Page) {
	    print $p->publish;
	}
	# last, pages
	my @Page = $this->get_all_pages(type=>"page",
					status=>"active",
					access=>"public",
					publish_method=>"static");
	my $found_index;
	foreach my $p (@Page) {
	    $share{Page} = $p;
	    print $p->publish;
	    my $page = $p->get;
	    if ($page->{filename} =~ /^(default|index).(php|html?)$/) {
		$found_index = $p;
	    }
	}
	if (! $found_index) {
	    $this->warn("Section::publish: no index page for section $section->{section_id}");
	    print "<p class=error>WARNING: no index page was found for this website.  
In most cases, it is recommended that you define an index page to serve 
as the front page of your site.  To set a particular page as the index page, 
edit that page's details, and set its filename to \"index.html\" 
(or \"index.php\" if using PHP pages).</p>\n";
	}
	# reset flush settings
	$| = $old_flush;
    }
    return;
}

1;


