#------------------------------------------------------------------------
# ExSite::Session
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
#
# An ExSite Session Manager
#
# A session is a hash of keys/values that is persistent between web
# site visits.  The session key is the address of the session, and is a
# hard-to-guess MD5 hash of login id, originating IP, and the current 
# time.  Old session keys are cleaned up by a garbage collector that 
# runs via cron.
#
# Sessions are stored in a DBM table for performance.  Because session 
# IDs are hard to guess, they obviate the necessity of revalidating 
# identities on each visit.
#
# We don't really care what version of DBM we use, since the entries
# are transient, anyway.
#
# Session information persists for a limited amount of time, assuming
# you run the garbage collector at regular intervals.  There are two
# different ways to expire a session:
#
# $config{session}->{max_lifetime} = max number of seconds since a
#                                    session was created.
# $config{session}->{max_delay}    = max number of seconds since a
#                                    session was last accessed.
#
# If either of these is exceeded, the session will be removed by the
# garbage collector.  max_delay defaults to 1 hour; max_lifetime defaults
# to 1 day.
#
# Sessions are useful for storing:
#   - user identity information
#   - account or website information
#   - current position in multi-page forms or reports
#   - modes and other state variables
#
# Pros of using sessions:
#   - helps to maintain state
#   - faster than re-authenticating on every transaction
# Cons of using sessions:
#   - extra infrastructure requirements (DBM session table)
#   - no persistent logins
#
# Methods:
#   make - create a session table
#   new - constructor
#   status - report session table stats
#   garbage_collector - clean out old sessions
#   
#   get - get session hash
#   set - set session hash
#   getval - get a session value
#   setval - set a session value
#   save - write session data back to the session table
#   close - drop the session hash
#   DESTROY - save and drop fallback
#   lock - lock the session table for saves
#   lock_nb - non-blocking version of lock
#   unlock - unlock the session table
#------------------------------------------------------------------------

package ExSite::Session;

use strict;
use ExSite::Config;
use ExSite::Misc;
use MD5;
use Fcntl qw(:flock);
use ExSite::Base;
our @ISA = qw(ExSite::Base);

# globals:

use vars qw($default_max_delay $default_max_lifetime);

$default_max_lifetime = 86400;  # maximum 1 day total session lifetime
$default_max_delay    = 3600;   # maximum 1 hour since last access

##### CLASS METHODS
# invoked like:
# $sess = new ExSite::Session;
# ExSite::Session->status;

# new: class constructor
#
# - gets session ID from HTTP_COOKIE
# - looks up session in session table, or creates a new one

sub new ($%) {
    my ($this,@data) = @_;

    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;

    # check for a session id in the cookie

    my $session_cookie = $config{site}->{id}."_sessionID";
    my $session_id = &get_cookie($session_cookie);
    my %session_data;

    # get access time

    my $time = time();

    my $session_file = $config{session}->{file} || "exsite.session";
    my %session_table;

    if (dbmopen(%session_table,$session_file,00666)) {
	if ($session_id) {  # user claims to have a session - must validate

	    $obj->info("re-opening session $session_id");
	    
	    # open session table
	    
	    if (exists $session_table{$session_id}) {
		
		# this is a good session
		
		%session_data = &DecodeString($session_table{$session_id});
		$session_data{'_atime'} = $time;
	    }
	    else {
		
		# no such session

		$obj->warn("session $session_id expired");
		dbmclose(%session_table);
		return undef;
	    }
	}
	else { # no session - create a new one
	    
	    # make an MD5 hash of the passed data
	    
	    if (@data == 0) {
		@data = ($ENV{REMOTE_ADDR}, time());
	    }
	    $session_id = MD5->hexhash(join("",@data));
	    $obj->info("creating new session $session_id");

	    # save the creation & access time

	    $session_data{'_ctime'} = $session_data{'_atime'} = $time;
	}
    }
    else {
	# no session table
	$obj->error("error opening session table");
	return undef;
    }

    # if we get here, we have declared a valid session;
    # save the DB handle, the ID, and the session data in the object

    $obj->{sessionDB} = \%session_table;
    $obj->{sessionID} = $session_id;
    $obj->{data} = \%session_data;
    return $obj;
}

# status report - display summary of active sessions

sub status {
    my ($this) = @_;
    my $time = time;
    my @sessions = keys %{$this->{sessionDB}};
    print scalar @sessions, " active sessions...\n";
    foreach my $sessionID (@sessions) {
	my %session = &DecodeString($this->{sessionDB}->{$sessionID});
	print "$sessionID: ",
	    $time - $session{_ctime}, " seconds old; ",
	    $time - $session{_atime}, " seconds since last access.\n";
    }
}

# garbage_collector - removes expired sessions
# sessions expire if more than 
# $config{session}->{max_lifetime} seconds elapsed since session creation
# $config{session}->{max_delay} seconds elapsed since last session activity
# - returns number of removed sessions

sub garbage_collector {
    my ($this) = @_;
    my $time = time();
    my @sessions = keys %{$this->{sessionDB}};
    my $lifetime = $config{session}->{max_lifetime} || $default_max_lifetime;
    my $latency = $config{session}->{max_delay} || $default_max_delay;
    my $gcount;
    foreach my $sessionID (@sessions) {
	my %session = &DecodeString($this->{sessionDB}->{$sessionID});
	if (($time > $session{_ctime} + $lifetime) ||
	    ($time > $session{_atime} + $latency)) {
	    delete $this->{sessionDB}->{$sessionID};
	    $gcount++;
	}
    }
    $this->info("Session::garbase_collector: removed $gcount sessions");
    return $gcount;
}

##### OBJECT METHODS - must first use new() to create an object
    
# get

sub get {
    my $this = shift;
    if (wantarray) { return %{$this->{data}}; }
    else { return $this->{data}; }
}

# set

sub set {
    my ($this,$rhash) = @_;
    # preserve session timestamps
    my $ctime = $this->{data}->{'_ctime'};
    my $atime = $this->{data}->{'_atime'};
    $this->{data} = $rhash;
    $this->{data}->{'_ctime'} = $ctime;
    $this->{data}->{'_atime'} = $atime;
}

# getval

sub getval {
    my ($this,$key) = @_;
    if (exists $this->{data}->{$key}) {
	return $this->{data}->{$key};
    }
    return undef;
}

# setval

sub setval {
    my ($this,$key,$val) = @_;
    $this->{data}->{$key} = $val;
}

# save - NB: can only save once per session

sub save {
    my ($this) = @_;
    if ($this->lock) {
	$this->{sessionDB}->{$this->{sessionID}} = &EncodeHash($this->{data});
	$this->unlock;
	dbmclose(%{$this->{sessionDB}});
	$this->{done} = 1;
	$this->info("session $this->{sessionID} updated");
	return 1;
    }
    else {
	$this->warn("session data NOT saved");
	return 0;
    }
}

# close

sub close {
    my $this = shift;
    delete $this->{sessionDB}->{$this->{sessionID}};
    dbmclose(%{$this->{sessionDB}});
    $this->{done} = 1;
}

# DESTROY - save the updated session, if necessary

sub DESTROY {
    my $this = shift;
    if (! $this->{done}) {
	$this->save;
    }
}

#------------------------------------------------------------------------
#
# auxilitary DBM tools
#
#------------------------------------------------------------------------

#
# Lock/unlock a DB for read-write access
# -> doesn't lock DB itself;  uses a lockfile for simplicity
#    will fail if lockfile is missing, or not writeable.
# -> returns TRUE (1) if lock was successful
# -> calling script must police itself - a misbehaving script
#    could ignore the lock and proceed anyway.
#

sub lock {
    my $this = shift;
    # arg0 is DBM file name, with path
    my $lockfile = $config{session}->{lockfile} || "session.lock";
    # first, test the lock file for write access
    if (open LOCK,$lockfile) {
	if (! flock(LOCK,LOCK_EX)) {
	    $this->warn("failed to lock session lockfile");
	    return 0;
	}
    }
    else {
	$this->warn("failed to open session lockfile");
	return 0;
    }
    return 1;
}

# non-blocking version - aborts after 10sec timeout
sub lock_nb {
    my $this = shift;
    my $lockfile = $config{session}->{lockfile} || "session.lock";
    # first, test the lock file for write access
    if (open LOCK,$lockfile) {
	my $count = 0;
	my $gotit = flock LOCK,(LOCK_EX|LOCK_NB);
	while (! $gotit && $count < 10) {
	    $count++;
	    sleep(1);
	    $gotit = flock LOCK,(LOCK_EX|LOCK_NB);
	}
        if (! $gotit) { 
	    $this->warn("session lockfile timed out");
	    return 0; 
	}
    }
    else { 
	$this->warn("failed to open session lockfile");
	return 0; 
    }
    return 1;
}

sub unlock {
    flock LOCK,LOCK_UN;
    close LOCK;
}

1;



















