package Modules::BaseDCD;
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------

=pod

BaseDCD : a base Dynamic Content Driver

BaseDCD gives a simple framework on which all Dynamic Content Drivers
should be based.  It may be inherited by any Dynamic Content Driver that
wishes to pick up its generic methods as a starting point for
development.

The driver framework is adapted from the Unix device driver API, which
is a simple I/O framework that is easy to understand, and which will
already be familiar to many programmers.

There are 3 standard methods used by all drivers:
    read: acquire data from an external source (such as the URL, stdin 
          [form input], a session manager, or database)
    write: output data for insertion into a web page
    ioctl: all operations that do not constitute either reads or writes
          (eg. user authentication, updating a log file, sending email)

Although DCDs may define other methods for internal use, only the above
three methods can be invoked directly by means of CMS tags.  The above
three methods may be thought of as the public driver methods.

The BaseDCD defines a set of generic DCD methods that can be
inherited by more advanced DCDs.  Generally speaking, the new()
constructor method is generic and may be reused by all DCDs.  The
read() method is fairly versatile, and should be adequate for parsing
most URLs and form input.  The write() method will almost certainly
need to be replaced with something appropriate to the DCD, as it is
the workhorse method in any CMS environment.  The ioctl() method,
while nominally useful here, serves more as an example of ioctl()
programming than as a useful piece of code.

=cut

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Input;

# inherit from Base class
use ExSite::Base;
use vars qw(@ISA);
@ISA = qw(ExSite::Base);

# new : called implicitly once per page when a DCM is invoked

sub new {
    my $this = shift;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->read;
    return $obj;
}

# read : called automatically by default new() to fetch input data
#        can optionally be invoked using <!--&Module.read(options)-->

sub read {
    my ($this,$opt) = @_;
    # fetch overlayed query and form input
    my $in = new ExSite::Input;
    $this->{input} = $in->combine;
}

# write : the default DCD call, invoked using <!--&BaseDCD(options)-->
#         or <!--&BaseDCD.write(options)-->  (these are synonymous).

sub write {
    my ($this,$options) = @_;
    # dummy routine
    return "BaseDCD dummy write method";
}

# ioctl : all operations that don't qualify as reads or writes
#         invoked as <!--BaseDCD.ioctl(request)-->

sub ioctl {
    my $this = shift;
    $_ = shift;   # $_ is the ioctl request
    # dummy routine - do nothing
    return 0;
}

# relink - recursive link back to same DCD, with different args
#          Returns a URL.

sub relink {
    my $this = shift;
    my $script = $ENV{SCRIPT_NAME};
    if ($ENV{PATH_INFO}) { $script .= "$ENV{PATH_INFO}"; }
    # go direct to query string, instead of via Input class, so that
    # we can pick up changes to $QUERY_STRING made after it was first
    # read
    my %query = &DecodeString($ENV{QUERY_STRING});
    my %newargs = (@_ == 1) ? %{$_[0]} : @_;
    foreach my $key (keys %newargs) {
	if (defined $newargs{$key}) {
	    $query{$key} = $newargs{$key};
	}
	else {
	    delete $query{$key};
	}
    }
    my $newquery = &EncodeHash(%query);
    
    return "$script?$newquery";
}

1;

